"""Core parametrizer classes for Tests."""

import inspect
import types

import pytest

import geomstats.backend as gs


def _is_isometry(
    metric,
    space,
    tangent_vec,
    trans_tangent_vec,
    base_point,
    is_tangent_atol,
    rtol,
    atol,
):
    """Check that a transformation is an isometry.

    This is an auxiliary function.

    Parameters
    ----------
    metric : RiemannianMetric
        Riemannian metric.
    tangent_vec : array-like
        Tangent vector at base point.
    trans_tangent_vec : array-like
        Transformed tangent vector at base point.
    base_point : array-like
        Point on manifold.
    is_tangent_atol: float
        Asbolute tolerance for the is_tangent function.
    rtol : float
        Relative tolerance to test this property.
    atol : float
        Absolute tolerance to test this property.
    """
    is_tangent = space.is_tangent(trans_tangent_vec, base_point, is_tangent_atol)
    is_equinormal = gs.isclose(
        metric.norm(trans_tangent_vec, base_point),
        metric.norm(tangent_vec, base_point),
        rtol,
        atol,
    )
    return gs.logical_and(is_tangent, is_equinormal)


class Parametrizer(type):
    """Metaclass for test files.

    Parametrizer decorates every function inside the class with pytest.mark.parametrizer
    (except class methods and static methods). Two conventions need to be respected:

    1.There should be a TestData object named 'testing_data'.
    2.Every test function should have its corresponding data function inside
    TestData object.

    Ex. test_exp() should have method called exp_data() inside 'testing_data'.
    """

    def __new__(cls, name, bases, attrs):
        for attr_name, attr_value in attrs.copy().items():
            if isinstance(attr_value, types.FunctionType):
                if ("skip_" + attr_name, True) not in locals()["attrs"].items():
                    args_str = ", ".join(inspect.getfullargspec(attr_value)[0][1:])
                    data_fn_str = attr_name[5:] + "_data"
                    attrs[attr_name] = pytest.mark.parametrize(
                        args_str,
                        getattr(locals()["attrs"]["testing_data"], data_fn_str)(),
                    )(attr_value)
                else:
                    del attrs[attr_name]

        return super(Parametrizer, cls).__new__(cls, name, bases, attrs)


class ManifoldParametrizer(Parametrizer):
    def __new__(cls, name, bases, attrs):
        def test_random_point_belongs(self, space_args, n_points, belongs_atol):
            """Check that a random point belongs to the manifold.

            Parameters
            ----------
            space_args : tuple
                Arguments to pass to constructor of the manifold.
            n_points : array-like
                Number of random points to sample.
            belongs_atol : float
                Absolute tolerance for the belongs function.
            """
            space = self.space(*space_args)
            random_point = space.random_point(n_points)
            result = gs.all(space.belongs(random_point, atol=belongs_atol))
            self.assertAllClose(result, True)

        def test_projection_belongs(self, space_args, point, belongs_atol):
            """Check that a point projected on a manifold belongs to the manifold.

            Parameters
            ----------
            space_args : tuple
                Arguments to pass to constructor of the manifold.
            point : array-like
                Point to be projected on the manifold.
            belongs_atol : float
                Absolute tolerance for the belongs function.
            """
            space = self.space(*space_args)
            belongs = space.belongs(space.projection(gs.array(point)), belongs_atol)
            self.assertAllClose(gs.all(belongs), gs.array(True))

        def test_to_tangent_is_tangent(
            self, space_args, vector, base_point, is_tangent_atol
        ):
            """Check that to_tangent returns a tangent vector.

            Parameters
            ----------
            space_args : tuple
                Arguments to pass to constructor of the manifold.
            vector : array-like
                Vector to be projected on the tangent space at base_point.
            base_point : array-like
                Point on the manifold.
            is_tangent_atol : float
                Absolute tolerance for the is_tangent function.
            """
            space = self.space(*space_args)
            tangent = space.to_tangent(gs.array(vector), gs.array(base_point))
            result = gs.all(
                space.is_tangent(tangent, gs.array(base_point), is_tangent_atol)
            )
            self.assertAllClose(result, gs.array(True))

        attrs[test_random_point_belongs.__name__] = test_random_point_belongs
        attrs[test_projection_belongs.__name__] = test_projection_belongs
        attrs[test_to_tangent_is_tangent.__name__] = test_to_tangent_is_tangent
        return super(ManifoldParametrizer, cls).__new__(cls, name, bases, attrs)


class OpenSetParametrizer(ManifoldParametrizer):
    def __new__(cls, name, bases, attrs):
        def test_to_tangent_is_tangent_in_ambient_space(
            self, space_args, vector, base_point, is_tangent_atol
        ):
            """Check that tangent vectors are in ambient space's tangent space.

            This projects a vector to the tangent space of the manifold, and
            then checks that tangent vector belongs to ambient space's tangent space.

            Parameters
            ----------
            space_args : tuple
                Arguments to pass to constructor of the manifold.
            vector : array-like
                Vector to be projected on the tangent space at base_point.
            base_point : array-like
                Point on the manifold.
            is_tangent_atol : float
                Absolute tolerance for the is_tangent function.
            """
            space = self.space(*space_args)
            tangent_vec = space.to_tangent(gs.array(vector), gs.array(base_point))
            result = gs.all(
                space.ambient_space.is_tangent(tangent_vec, is_tangent_atol)
            )
            self.assertAllClose(result, gs.array(True))

        attrs[
            test_to_tangent_is_tangent_in_ambient_space.__name__
        ] = test_to_tangent_is_tangent_in_ambient_space
        return super(OpenSetParametrizer, cls).__new__(cls, name, bases, attrs)


class LieGroupParametrizer(ManifoldParametrizer):
    def __new__(cls, name, bases, attrs):
        def test_exp_log_composition(
            self, group_args, tangent_vec, base_point, rtol, atol
        ):
            """Check that group exponential and logarithm are inverse.

            This is calling group exponential first, then group logarithm.

            Parameters
            ----------
            group_args : tuple
                Arguments to pass to constructor of the group.
            tangent_vec : array-like
                Tangent vector to the manifold at base_point.
            base_point : array-like
                Point on the manifold.
            rtol : float
                Relative tolerance to test this property.
            atol : float
                Absolute tolerance to test this property.
            """
            group = self.group(*group_args)
            exp_point = group.exp(gs.array(tangent_vec), gs.array(base_point))
            log_vec = group.log(exp_point, gs.array(base_point))
            self.assertAllClose(log_vec, gs.array(tangent_vec), rtol, atol)

        def test_log_exp_composition(self, group_args, point, base_point, rtol, atol):
            """Check that group exponential and logarithm are inverse.

            This is calling group logarithm first, then group exponential.

            Parameters
            ----------
            group_args : tuple
                Arguments to pass to constructor of the group.
            point : array-like
                Point on the manifold.
            base_point : array-like
                Point on the manifold.
            rtol : float
                Relative tolerance to test this property.
            atol : float
                Absolute tolerance to test this property.
            """
            group = self.group(*group_args)
            log_vec = group.log(gs.array(point), gs.array(base_point))
            exp_point = group.exp(log_vec, gs.array(base_point))
            self.assertAllClose(exp_point, gs.array(point), rtol, atol)

        attrs[test_exp_log_composition.__name__] = test_exp_log_composition
        attrs[test_log_exp_composition.__name__] = test_log_exp_composition
        return super(LieGroupParametrizer, cls).__new__(cls, name, bases, attrs)


class VectorSpaceParametrizer(ManifoldParametrizer):
    def __new__(cls, name, bases, attrs):
        def test_basis_belongs(self, space_args, belongs_atol):
            """Check that basis elements belong to vector space.

            Parameters
            ----------
            space_args : tuple
                Arguments to pass to constructor of the vector space.
            belongs_atol : float
                Absolute tolerance of the belongs function.
            """
            space = self.space(*space_args)
            result = gs.all(space.belongs(space.basis, belongs_atol))
            self.assertAllClose(result, gs.array(True))

        def test_basis_cardinality(self, space_args):
            """Check that the number of basis elements is the dimension.

            Parameters
            ----------
            space_args : tuple
                Arguments to pass to constructor of the vector space.
            """
            space = self.space(*space_args)
            basis = space.basis
            self.assertAllClose(len(basis), space.dim)

        attrs[test_basis_belongs.__name__] = test_basis_belongs
        attrs[test_basis_cardinality.__name__] = test_basis_cardinality
        return super(VectorSpaceParametrizer, cls).__new__(cls, name, bases, attrs)


class MatrixLieAlgebraParametrizer(VectorSpaceParametrizer):
    def __new__(cls, name, bases, attrs):
        def test_basis_representation_matrix_representation_composition(
            self, algebra_args, matrix_rep, rtol, atol
        ):
            """Check that changing coordinate system twice gives back the point.

            A point written in basis representation is converted to matrix
            representation and back.

            Parameters
            ----------
            algebra_args : tuple
                Arguments to pass to constructor of the Lie algebra.
            matrix_rep : array-like
                Point on the Lie algebra given in its matrix representation.
            rtol : float
                Relative tolerance to test this property.
            atol : float
                Absolute tolerance to test this property.
            """
            algebra = self.algebra(*algebra_args)
            basis_rep = algebra.basis_representation(gs.array(matrix_rep))
            result = algebra.matrix_representation(basis_rep)
            self.assertAllClose(result, gs.array(matrix_rep), rtol, atol)

        def test_matrix_representation_basis_representation_composition(
            self,
            algebra_args,
            basis_rep,
            rtol,
            atol,
        ):
            """Check that changing coordinate system twice gives back the point.

            A point written in matrix representation is converted to basis
            representation and back.

            Parameters
            ----------
            algebra_args : tuple
                Arguments to pass to constructor of the Lie algebra.
            basis_rep : array-like
                Point on the Lie algebra given in its basis representation.
            rtol : float
                Relative tolerance to test this property.
            atol : float
                Absolute tolerance to test this property.
            """
            algebra = self.algebra(*algebra_args)
            mat_rep = algebra.matrix_representation(basis_rep)
            result = algebra.basis_representation(mat_rep)
            self.assertAllClose(result, gs.array(basis_rep), rtol, atol)

        attrs[
            test_basis_representation_matrix_representation_composition.__name__
        ] = test_basis_representation_matrix_representation_composition
        attrs[
            test_matrix_representation_basis_representation_composition.__name__
        ] = test_matrix_representation_basis_representation_composition
        return super(MatrixLieAlgebraParametrizer, cls).__new__(cls, name, bases, attrs)


class LevelSetParametrizer(ManifoldParametrizer):
    def __new__(cls, name, bases, attrs):
        def test_extrinsic_intrinsic_composition(
            self, space_args, point_extrinsic, rtol, atol
        ):
            """Check that changing coordinate system twice gives back the point.

            A point written in extrinsic coordinates is converted to
            intrinsic coordinates and back.

            Parameters
            ----------
            space_args : tuple
                Arguments to pass to constructor of the manifold.
            point_extrinsic : array-like
                Point on the manifold in extrinsic coordinates.
            rtol : float
                Relative tolerance to test this property.
            atol : float
                Absolute tolerance to test this property.
            """
            space = self.space(*space_args)
            point_intrinsic = space.extrinsic_to_intrinsic_coords(point_extrinsic)
            result = space.intrinsic_to_extrinsic_coords(point_intrinsic)
            expected = point_extrinsic
            self.assertAllClose(result, expected, rtol, atol)

        def test_intrinsic_extrinsic_composition(
            self, space_args, point_intrinsic, rtol, atol
        ):
            """Check that changing coordinate system twice gives back the point.

            A point written in intrinsic coordinates is converted to
            extrinsic coordinates and back.

            Parameters
            ----------
            space_args : tuple
                Arguments to pass to constructor of the manifold.
            point_intrinsic : array-like
                Point on the manifold in intrinsic coordinates.
            rtol : float
                Relative tolerance to test this property.
            atol : float
                Absolute tolerance to test this property.
            """
            space = self.space(*space_args)
            point_extrinsic = space.intrinsic_to_extrinsic_coords(point_intrinsic)
            result = space.extrinsic_to_intrinsic_coords(point_extrinsic)
            expected = point_intrinsic

            self.assertAllClose(result, expected, rtol, atol)

        attrs[
            test_extrinsic_intrinsic_composition.__name__
        ] = test_extrinsic_intrinsic_composition
        attrs[
            test_intrinsic_extrinsic_composition.__name__
        ] = test_intrinsic_extrinsic_composition
        return super(LevelSetParametrizer, cls).__new__(cls, name, bases, attrs)


class ConnectionParametrizer(Parametrizer):
    def __new__(cls, name, bases, attrs):
        def test_exp_shape(self, connection_args, tangent_vec, base_point, expected):
            """Check that exp returns an array of the expected shape.

            Parameters
            ----------
            connection_args : tuple
                Arguments to pass to constructor of the connection.
            tangent_vec : array-like
                Tangent vector at base point.
            base_point : array-like
                Point on the manifold.
            expected : tuple
                Expected shape for the result of the exp function.
            """
            connection = self.connection(*connection_args)
            exp = connection.exp(gs.array(tangent_vec), gs.array(base_point))
            result = gs.shape(exp)
            self.assertAllClose(result, expected)

        def test_log_shape(self, connection_args, point, base_point, expected):
            """Check that log returns an array of the expected shape.

            Parameters
            ----------
            connection_args : tuple
                Arguments to pass to constructor of the connection.
            point : array-like
                Point on the manifold.
            base_point : array-like
                Point on the manifold.
            expected : tuple
                Expected shape for the result of the log function.
            """
            connection = self.connection(*connection_args)
            log = connection.log(gs.array(point), gs.array(base_point))
            result = gs.shape(log)
            self.assertAllClose(result, expected)

        def test_exp_belongs(
            self, connection_args, space, tangent_vec, base_point, belongs_atol
        ):
            """Check that the connection exponential gives a point on the manifold.

            Parameters
            ----------
            connection_args : tuple
                Arguments to pass to constructor of the connection.
            space : Manifold
                Manifold where connection is defined.
            tangent_vec : array-like
                Tangent vector at base point.
            base_point : array-like
                Point on the manifold.
            belongs_atol : float
                Absolute tolerance for the belongs function.
            """
            connection = self.connection(*connection_args)
            exp = connection.exp(gs.array(tangent_vec), gs.array(base_point))
            result = gs.all(space.belongs(exp, belongs_atol))
            self.assertAllClose(result, gs.array(True))

        def test_log_is_tangent(
            self, connection_args, space, point, base_point, is_tangent_atol
        ):
            """Check that the connection logarithm gives a tangent vector.

            Parameters
            ----------
            connection_args : tuple
                Arguments to pass to constructor of the connection.
            space : Manifold
                Manifold where connection is defined.
            point : array-like
                Point on the manifold.
            base_point : array-like
                Point on the manifold.
            is_tangent_atol : float
                Absolute tolerance for the is_tangent function.
            """
            connection = self.connection(*connection_args)
            log = connection.log(gs.array(point), gs.array(base_point))
            result = gs.all(
                space.is_tangent(log, gs.array(base_point), is_tangent_atol)
            )
            self.assertAllClose(result, gs.array(True))

        def test_geodesic_ivp_belongs(
            self,
            connection_args,
            space,
            n_points,
            initial_point,
            initial_tangent_vec,
            belongs_atol,
        ):
            """Check that connection geodesics belong to manifold.

            This is for geodesics defined by the initial value problem (ivp).

            Parameters
            ----------
            connection_args : tuple
                Arguments to pass to constructor of the connection.
            space : Manifold
                Manifold where connection is defined.
            n_points : int
                Number of points on the geodesics.
            initial_point : array-like
                Point on the manifold.
            initial_tangent_vec : array-like
                Tangent vector at base point.
            belongs_atol : float
                Absolute tolerance for the belongs function.
            """
            connection = self.connection(*connection_args)
            geodesic = connection.geodesic(
                initial_point=initial_point, initial_tangent_vec=initial_tangent_vec
            )

            t = gs.linspace(start=0.0, stop=1.0, num=n_points)
            points = geodesic(t)

            result = space.belongs(points, belongs_atol)
            expected = gs.array(n_points * [True])

            self.assertAllClose(result, expected)

        def test_geodesic_bvp_belongs(
            self,
            connection_args,
            space,
            n_points,
            initial_point,
            end_point,
            belongs_atol,
        ):
            """Check that connection geodesics belong to manifold.

            This is for geodesics defined by the boundary value problem (bvp).

            Parameters
            ----------
            connection_args : tuple
                Arguments to pass to constructor of the connection.
            space : Manifold
                Manifold where connection is defined.
            n_points : int
                Number of points on the geodesics.
            initial_point : array-like
                Point on the manifold.
            end_point : array-like
                Point on the manifold.
            belongs_atol : float
                Absolute tolerance for the belongs function.
            """
            connection = self.connection(*connection_args)

            geodesic = connection.geodesic(
                initial_point=initial_point, end_point=end_point
            )

            t = gs.linspace(start=0.0, stop=1.0, num=n_points)
            points = geodesic(t)

            result = space.belongs(points, belongs_atol)
            expected = gs.array(n_points * [True])

            self.assertAllClose(result, expected)

        def test_log_exp_composition(
            self, connection_args, point, base_point, rtol, atol
        ):
            """Check that connection logarithm and exponential are inverse.

            This is calling connection logarithm first, then connection exponential.

            Parameters
            ----------
            connection_args : tuple
                Arguments to pass to constructor of the connection.
            point : array-like
                Point on the manifold.
            base_point : array-like
                Point on the manifold.
            rtol : float
                Relative tolerance to test this property.
            atol : float
                Absolute tolerance to test this property.
            """
            connection = self.connection(*connection_args)
            log = connection.log(gs.array(point), base_point=gs.array(base_point))
            result = connection.exp(tangent_vec=log, base_point=gs.array(base_point))
            self.assertAllClose(result, point, rtol=rtol, atol=atol)

        def test_exp_log_composition(
            self, connection_args, tangent_vec, base_point, rtol, atol
        ):
            """Check that connection exponential and logarithm are inverse.

            This is calling connection exponential first, then connection logarithm.

            Parameters
            ----------
            connection_args : tuple
                Arguments to pass to constructor of the connection.
            tangent_vec : array-like
                Tangent vector to the manifold at base_point.
            base_point : array-like
                Point on the manifold.
            rtol : float
                Relative tolerance to test this property.
            atol : float
                Absolute tolerance to test this property.
            """
            connection = self.connection(*connection_args)
            exp = connection.exp(
                tangent_vec=tangent_vec, base_point=gs.array(base_point)
            )
            result = connection.log(exp, base_point=gs.array(base_point))
            self.assertAllClose(result, tangent_vec, rtol=rtol, atol=atol)

        def test_exp_ladder_parallel_transport(
            self,
            connection_args,
            direction,
            tangent_vec,
            base_point,
            scheme,
            n_rungs,
            alpha,
            rtol,
            atol,
        ):
            """Check that end point of ladder parallel transport matches exponential.

            Parameters
            ----------
            connection_args : tuple
                Arguments to pass to constructor of the connection.
            direction : array-like
                Tangent vector to the manifold at base_point.
            tangent_vec : array-like
                Tangent vector to the manifold at base_point.
            base_point : array-like
                Point on the manifold.
            scheme : str, {'pole', 'schild'}
                The scheme to use for the construction of the ladder at each step.
            n_rungs : int
                Number of steps of the ladder.
            alpha : float
                Exponent for the scaling of the vector to transport.
            rtol : float
                Relative tolerance to test this property.
            atol : float
                Absolute tolerance to test this property.
            """
            connection = self.connection(*connection_args)

            ladder = connection.ladder_parallel_transport(
                tangent_vec,
                base_point,
                direction,
                n_rungs=n_rungs,
                scheme=scheme,
                alpha=alpha,
            )

            result = ladder["end_point"]
            expected = connection.exp(direction, base_point)

            self.assertAllClose(result, expected, rtol=rtol, atol=atol)

        def test_exp_geodesic_ivp(
            self, connection_args, n_points, tangent_vec, base_point, rtol, atol
        ):
            """Check that end point of geodesic matches exponential.

            Parameters
            ----------
            connection_args : tuple
                Arguments to pass to constructor of the connection.
            n_points : int
                Number of points on the geodesic.
            tangent_vec : array-like
                Tangent vector to the manifold at base_point.
            base_point : array-like
                Point on the manifold.
            rtol : float
                Relative tolerance to test this property.
            atol : float
                Absolute tolerance to test this property.
            """
            connection = self.connection(*connection_args)
            geodesic = connection.geodesic(
                initial_point=base_point, initial_tangent_vec=tangent_vec
            )
            t = gs.linspace(start=0.0, stop=1.0, num=n_points)
            points = geodesic(t)
            result = points[:, -1]
            expected = connection.exp(tangent_vec, base_point)
            self.assertAllClose(expected, result, rtol=rtol, atol=atol)

        attrs[test_exp_shape.__name__] = test_exp_shape
        attrs[test_log_shape.__name__] = test_log_shape
        attrs[test_exp_belongs.__name__] = test_exp_belongs
        attrs[test_log_is_tangent.__name__] = test_log_is_tangent
        attrs[test_geodesic_ivp_belongs.__name__] = test_geodesic_ivp_belongs
        attrs[test_geodesic_bvp_belongs.__name__] = test_geodesic_bvp_belongs
        attrs[test_log_exp_composition.__name__] = test_log_exp_composition
        attrs[test_exp_log_composition.__name__] = test_exp_log_composition
        attrs[
            test_exp_ladder_parallel_transport.__name__
        ] = test_exp_ladder_parallel_transport
        attrs[test_exp_geodesic_ivp.__name__] = test_exp_geodesic_ivp

        return super(ConnectionParametrizer, cls).__new__(cls, name, bases, attrs)


class RiemannianMetricParametrizer(ConnectionParametrizer):
    def __new__(cls, name, bases, attrs):
        def test_squared_dist_is_symmetric(
            self, metric_args, point_a, point_b, rtol, atol
        ):
            """Check that the squared geodesic distance is symmetric.

            Parameters
            ----------
            metric_args : tuple
                Arguments to pass to constructor of the metric.
            point_a : array-like
                Point on the manifold.
            point_b : array-like
                Point on the manifold.
            rtol : float
                Relative tolerance to test this property.
            atol : float
                Absolute tolerance to test this property.
            """
            metric = self.metric(*metric_args)
            sd_a_b = metric.squared_dist(gs.array(point_a), gs.array(point_b))
            sd_b_a = metric.squared_dist(gs.array(point_b), gs.array(point_a))
            self.assertAllClose(sd_a_b, sd_b_a, rtol=rtol, atol=atol)

        def test_parallel_transport_ivp_is_isometry(
            self,
            metric_args,
            space,
            tangent_vec,
            base_point,
            direction,
            is_tangent_atol,
            rtol,
            atol,
        ):
            """Check that parallel transport is an isometry.

            This is for parallel transport defined by initial value problem (ivp).

            Parameters
            ----------
            metric_args : tuple
                Arguments to pass to constructor of the metric.
            space : Manifold
                Manifold where metric is defined.
            tangent_vec : array-like
                Tangent vector at base point, to be transported.
            base_point : array-like
                Point on manifold.
            direction : array-like
                Tangent vector at base point, along which to transport.
            is_tangent_atol: float
                Asbolute tolerance for the is_tangent function.
            rtol : float
                Relative tolerance to test this property.
            atol : float
                Absolute tolerance to test this property.
            """
            metric = self.metric(*metric_args)

            end_point = metric.exp(direction, base_point)

            transported = metric.parallel_transport(tangent_vec, base_point, direction)
            result = _is_isometry(
                metric,
                space,
                tangent_vec,
                transported,
                end_point,
                is_tangent_atol,
                rtol,
                atol,
            )
            expected = gs.array(len(result) * [True])
            self.assertAllClose(result, expected)

        def test_parallel_transport_bvp_is_isometry(
            self,
            metric_args,
            space,
            tangent_vec,
            base_point,
            end_point,
            is_tangent_atol,
            rtol,
            atol,
        ):
            """Check that parallel transport is an isometry.

            This is for parallel transport defined by boundary value problem (ivp).

            Parameters
            ----------
            metric_args : tuple
                Arguments to pass to constructor of the metric.
            space : Manifold
                Manifold where metric is defined.
            tangent_vec : array-like
                Tangent vector at base point, to be transported.
            base_point : array-like
                Point on manifold.
            end_point : array-like
                Point on manifold.
            is_tangent_atol: float
                Asbolute tolerance for the is_tangent function.
            rtol : float
                Relative tolerance to test this property.
            atol : float
                Absolute tolerance to test this property.
            """
            metric = self.metric(*metric_args)

            transported = metric.parallel_transport(
                tangent_vec, base_point, end_point=end_point
            )
            result = _is_isometry(
                metric,
                space,
                tangent_vec,
                transported,
                end_point,
                is_tangent_atol,
                rtol,
                atol,
            )
            expected = gs.array(len(result) * [True])
            self.assertAllClose(result, expected)

        attrs[test_squared_dist_is_symmetric.__name__] = test_squared_dist_is_symmetric
        attrs[
            test_parallel_transport_ivp_is_isometry.__name__
        ] = test_parallel_transport_ivp_is_isometry
        attrs[
            test_parallel_transport_bvp_is_isometry.__name__
        ] = test_parallel_transport_bvp_is_isometry
        return super(RiemannianMetricParametrizer, cls).__new__(cls, name, bases, attrs)



"""Pre-process time series into batched covariance matrices.

The user defines the number of time steps of the batches.
It starts by removing the transient signal by taking a margin on each side
of the sign change. It then creates batches of data that will be used to
build the covariance matrices. In practice, one needs to choose the size
of the batches big enough to get enough information, and small enough so
that the online classifier is reactive enough.

Lead author: Marius Guerard.
"""

import numpy as np

import geomstats.backend as gs
from geomstats.geometry.symmetric_matrices import SymmetricMatrices


class TimeSeriesCovariance:
    """Class for generating a list of covariance matrices from time series.

    Prepare a TimeSeriesCovariance Object from time series in dictionary.

    Parameters
    ----------
    data_dict : dict
        Dictionary with 'time', 'raw_data', 'label' as key
        and the corresponding array as values.
    n_steps : int
        Size of the batches.
    n_timeseries : int
        The number of electrodes used for the recording.
    label_map : dictionary
        Encode the label into digits.
    margin : int
        Number of index to remove before and after a sign change (Can
        help getting a stationary signal).

    Attributes
    ----------
    label_map : dictionary
        Encode the label into digits.
    data_dict : dict
        Dictionary with 'time', 'raw_data', 'label' as key
        and the corresponding array as values.
    n_steps : int
        Size of the batches.
    n_timeseries : int
        The number of electrodes used for the recording.
    batches : array
        The start indexes of the batches to use to compute covariance matrices.
    margin : int
        Number of index to remove before and after a sign change (Can
        help getting a stationary signal).
    covs : array
        The covariance matrices.
    labels : array
        The digit labels corresponding to each batch.
    covec : array
        The vectorized version of the covariance matrices.
    diags : array
        The covariance matrices diagonals.
    """

    def __init__(self, data, n_steps, n_timeseries, label_map, margin=0):
        self.label_map = label_map
        self.data = data
        self.n_steps = n_steps
        self.n_timeseries = n_timeseries
        self.batches = gs.array([])
        self.margin = margin
        self.covs = gs.array([])
        self.labels = gs.array([])
        self.covecs = gs.array([])
        self.diags = gs.array([])

    def _format_labels(self):
        """Convert the labels into digits."""
        self.data["y"] = gs.array([self.label_map[x] for x in self.data["label"]])

    def _create_batches(self):
        """Create the batches used to compute covariance matrices.

        If margin != 0, we add an index margin at each label change
        to get stationary signal corresponding to each label.
        """
        start_ids = gs.where(np.diff(self.data["y"]) != 0)[0]
        end_ids = np.append(start_ids[1:], len(self.data)) - self.margin
        start_ids += self.margin
        batches_list = [
            range(start_id, end_id - self.n_steps, self.n_steps)
            for start_id, end_id in zip(start_ids, end_ids)
        ]
        self.batches = np.int_(gs.concatenate(batches_list))

    def transform(self):
        """Transform the time series into batched covariance matrices.

        We also compute the corresponding vectors, variance vector,
        labels, and experiments.
        """
        if "y" not in self.data.keys():
            self._format_labels()
        self._create_batches()
        covs = []
        for i in self.batches:
            x = self.data["raw_data"][i : i + self.n_steps]
            covs.append(np.cov(x.transpose()))
        self.labels = gs.array(self.data["y"][self.batches])
        self.covs = gs.array(covs)
        self.covecs = gs.array([SymmetricMatrices.to_vector(cov) for cov in self.covs])
        self.diags = self.covs.diagonal(0, 1, 2)



"""Unit tests for Positive lower triangular matrices"""


import math
import random

import geomstats.backend as gs
import geomstats.tests
from geomstats.geometry.positive_lower_triangular_matrices import (
    CholeskyMetric,
    PositiveLowerTriangularMatrices,
)
from geomstats.geometry.symmetric_matrices import SymmetricMatrices
from tests.conftest import TestCase
from tests.data_generation import OpenSetTestData, RiemannianMetricTestData
from tests.parametrizers import OpenSetParametrizer, RiemannianMetricParametrizer

EULER = gs.exp(1.0)
SQRT_2 = math.sqrt(2)


class TestPositiveLowerTriangularMatrices(TestCase, metaclass=OpenSetParametrizer):
    """Test of Cholesky methods."""

    space = PositiveLowerTriangularMatrices

    class TestDataPositiveLowerTriangularMatrices(OpenSetTestData):
        n_list = random.sample(range(2, 5), 2)
        space_args_list = [(n,) for n in n_list]
        shape_list = [(n, n) for n in n_list]
        n_samples_list = random.sample(range(2, 5), 2)
        n_points_list = random.sample(range(2, 5), 2)
        n_vecs_list = random.sample(range(2, 5), 2)

        def belongs_data(self):
            smoke_data = [
                dict(n=2, mat=[[1.0, 0.0], [-1.0, 3.0]], expected=True),
                dict(n=2, mat=[[1.0, -1.0], [-1.0, 3.0]], expected=False),
                dict(n=2, mat=[[-1.0, 0.0], [-1.0, 3.0]], expected=False),
                dict(n=3, mat=[[1.0, 0], [0, 1.0]], expected=False),
                dict(
                    n=2,
                    mat=[
                        [[1.0, 0], [0, 1.0]],
                        [[1.0, 2.0], [2.0, 1.0]],
                        [[-1.0, 0.0], [1.0, 1.0]],
                        [[0.0, 0.0], [1.0, 1.0]],
                    ],
                    expected=[True, False, False, False],
                ),
                dict(
                    n=3,
                    mat=[
                        [[1.0, 0.0, 1.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
                        [[0.0, 0.0, 0.0], [2.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
                        [[1.0, 0.0, 0.0], [2.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
                        [[-1.0, 0.0, 0.0], [2.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
                    ],
                    expected=[False, False, True, False],
                ),
            ]
            return self.generate_tests(smoke_data)

        def random_point_and_belongs_data(self):
            smoke_data = [
                dict(n=1, n_samples=1),
                dict(n=2, n_samples=2),
                dict(n=10, n_samples=100),
                dict(n=100, n_samples=10),
            ]
            return self.generate_tests(smoke_data)

        def gram_data(self):
            smoke_data = [
                dict(
                    n=2,
                    point=[[1.0, 0.0], [2.0, 1.0]],
                    expected=[[1.0, 2.0], [2.0, 5.0]],
                ),
                dict(
                    n=2,
                    point=[[[2.0, 1.0], [0.0, 1.0]], [[-6.0, 0.0], [5.0, 3.0]]],
                    expected=[[[5.0, 1.0], [1.0, 1.0]], [[36.0, -30.0], [-30.0, 34.0]]],
                ),
            ]
            return self.generate_tests(smoke_data)

        def differential_gram_data(self):
            smoke_data = [
                dict(
                    n=2,
                    tangent_vec=[[-1.0, 0.0], [2.0, -1.0]],
                    base_point=[[1.0, 0.0], [2.0, 1.0]],
                    expected=[[-2.0, 0.0], [0.0, 6.0]],
                ),
                dict(
                    n=2,
                    tangent_vec=[[[-1.0, 2.0], [2.0, -1.0]], [[0.0, 4.0], [4.0, -1.0]]],
                    base_point=[[[3.0, 0.0], [-1.0, 2.0]], [[4.0, 0.0], [-1.0, 4.0]]],
                    expected=[
                        [[-6.0, 11.0], [11.0, -8.0]],
                        [[0.0, 32.0], [32.0, -16.0]],
                    ],
                ),
            ]
            return self.generate_tests(smoke_data)

        def inverse_differential_gram_data(self):
            smoke_data = [
                dict(
                    n=2,
                    tangent_vec=[[1.0, 2.0], [2.0, 5.0]],
                    base_point=[[1.0, 0.0], [2.0, 2.0]],
                    expected=[[0.5, 0.0], [1.0, 0.25]],
                ),
                dict(
                    n=2,
                    tangent_vec=[[[-4.0, 1.0], [1.0, -4.0]], [[0.0, 4.0], [4.0, -8.0]]],
                    base_point=[[[2.0, 0.0], [-1.0, 2.0]], [[4.0, 0.0], [-1.0, 2.0]]],
                    expected=[[[-1.0, 0.0], [0.0, -1.0]], [[0.0, 0.0], [1.0, -1.5]]],
                ),
            ]
            return self.generate_tests(smoke_data)

        def differential_gram_belongs_data(self):
            n_list = [1, 2, 2, 3, 10, 5, 100, 200]
            n_samples_list = [1, 1, 2, 50, 1000, 100, 10, 5]
            space = PositiveLowerTriangularMatrices
            random_data = [
                dict(
                    n=n,
                    tangent_vec=space(n).ambient_space.random_point(n_samples),
                    base_point=space(n).random_point(n_samples),
                )
                for n, n_samples in zip(n_list, n_samples_list)
            ]
            return self.generate_tests([], random_data)

        def inverse_differential_gram_belongs_data(self):
            n_list = [1, 2, 2, 3, 10, 15, 100, 200]
            n_samples_list = [1, 1, 2, 200, 1000, 100, 10, 5]
            space = PositiveLowerTriangularMatrices
            random_data = [
                dict(
                    n=n,
                    tangent_vec=space(n).ambient_space.random_point(n_samples),
                    base_point=space(n).random_point(n_samples),
                )
                for n, n_samples in zip(n_list, n_samples_list)
            ]
            return self.generate_tests([], random_data)

        def random_point_belongs_data(self):
            smoke_space_args_list = [(2,), (3,)]
            smoke_n_points_list = [1, 2]
            return self._random_point_belongs_data(
                smoke_space_args_list,
                smoke_n_points_list,
                self.space_args_list,
                self.n_points_list,
            )

        def projection_belongs_data(self):
            return self._projection_belongs_data(
                self.space_args_list, self.shape_list, self.n_samples_list
            )

        def to_tangent_is_tangent_data(self):
            return self._to_tangent_is_tangent_data(
                PositiveLowerTriangularMatrices,
                self.space_args_list,
                self.shape_list,
                self.n_vecs_list,
            )

        def to_tangent_is_tangent_in_ambient_space_data(self):
            return self._to_tangent_is_tangent_in_ambient_space_data(
                PositiveLowerTriangularMatrices, self.space_args_list, self.shape_list
            )

    testing_data = TestDataPositiveLowerTriangularMatrices()

    def test_belongs(self, n, mat, expected):
        self.assertAllClose(self.space(n).belongs(gs.array(mat)), gs.array(expected))

    def test_gram(self, n, point, expected):
        self.assertAllClose(self.space(n).gram(gs.array(point)), gs.array(expected))

    def test_differential_gram(self, n, tangent_vec, base_point, expected):
        self.assertAllClose(
            self.space(n).differential_gram(
                gs.array(tangent_vec), gs.array(base_point)
            ),
            gs.array(expected),
        )

    def test_inverse_differential_gram(self, n, tangent_vec, base_point, expected):
        self.assertAllClose(
            self.space(n).inverse_differential_gram(
                gs.array(tangent_vec), gs.array(base_point)
            ),
            gs.array(expected),
        )

    @geomstats.tests.np_and_autograd_only
    def test_differential_gram_belongs(self, n, tangent_vec, base_point):
        result = self.space(n).differential_gram(
            gs.array(tangent_vec), gs.array(base_point)
        )
        self.assertAllClose(gs.all(SymmetricMatrices(n).belongs(result)), True)

    def test_inverse_differential_gram_belongs(self, n, tangent_vec, base_point):
        result = self.space(n).inverse_differential_gram(
            gs.array(tangent_vec), gs.array(base_point)
        )
        self.assertAllClose(gs.all(self.space(n).ambient_space.belongs(result)), True)


class TestCholeskyMetric(TestCase, metaclass=RiemannianMetricParametrizer):
    metric = connection = CholeskyMetric
    space = PositiveLowerTriangularMatrices

    skip_test_parallel_transport_ivp_is_isometry = True
    skip_test_parallel_transport_bvp_is_isometry = True
    skip_test_exp_geodesic_ivp = True

    class TestDataCholeskyMetric(RiemannianMetricTestData):
        n_list = random.sample(range(2, 7), 5)
        metric_args_list = [(n,) for n in n_list]
        shape_list = [(n, n) for n in n_list]
        space_list = [PositiveLowerTriangularMatrices(n) for n in n_list]
        n_points_list = random.sample(range(1, 7), 5)
        n_samples_list = random.sample(range(1, 7), 5)
        n_points_a_list = random.sample(range(1, 7), 5)
        n_points_b_list = [1]
        batch_size_list = random.sample(range(2, 7), 5)
        alpha_list = [1] * 5
        n_rungs_list = [1] * 5
        scheme_list = ["pole"] * 5

        def diag_inner_product_data(self):
            smoke_data = [
                dict(
                    n=2,
                    tangent_vec_a=[[1.0, 0.0], [-2.0, -1.0]],
                    tangent_vec_b=[[2.0, 0.0], [-3.0, -1.0]],
                    base_point=[[SQRT_2, 0.0], [-3.0, 1.0]],
                    expected=2.0,
                )
            ]
            return self.generate_tests(smoke_data)

        def strictly_lower_inner_product_data(self):
            smoke_data = [
                dict(
                    n=2,
                    tangent_vec_a=[[1.0, 0.0], [-2.0, -1.0]],
                    tangent_vec_b=[[2.0, 0.0], [-3.0, -1.0]],
                    expected=6.0,
                )
            ]
            return self.generate_tests(smoke_data)

        def inner_product_data(self):
            smoke_data = [
                dict(
                    n=2,
                    tangent_vec_a=[[1.0, 0.0], [-2.0, -1.0]],
                    tangent_vec_b=[[2.0, 0.0], [-3.0, -1.0]],
                    base_point=[[SQRT_2, 0.0], [-3.0, 1.0]],
                    expected=8.0,
                ),
                dict(
                    n=2,
                    tangent_vec_a=[
                        [[3.0, 0.0], [4.0, 2.0]],
                        [[-1.0, 0.0], [2.0, -4.0]],
                    ],
                    tangent_vec_b=[[[4.0, 0.0], [3.0, 3.0]], [[3.0, 0.0], [-6.0, 2.0]]],
                    base_point=[[[3, 0.0], [-2.0, 6.0]], [[1, 0.0], [-1.0, 1.0]]],
                    expected=[13.5, -23.0],
                ),
            ]
            return self.generate_tests(smoke_data)

        def exp_data(self):
            smoke_data = [
                dict(
                    n=2,
                    tangent_vec=[[-1.0, 0.0], [2.0, 3.0]],
                    base_point=[[1.0, 0.0], [2.0, 2.0]],
                    expected=[[1 / EULER, 0.0], [4.0, 2 * gs.exp(1.5)]],
                ),
                dict(
                    n=2,
                    tangent_vec=[[[0.0, 0.0], [2.0, 0.0]], [[1.0, 0.0], [0.0, 0.0]]],
                    base_point=[[[1.0, 0.0], [2.0, 2.0]], [[1.0, 0.0], [0.0, 2.0]]],
                    expected=[
                        [[1.0, 0.0], [4.0, 2.0]],
                        [[gs.exp(1.0), 0.0], [0.0, 2.0]],
                    ],
                ),
            ]
            return self.generate_tests(smoke_data)

        def log_data(self):
            smoke_data = [
                dict(
                    n=2,
                    point=[[EULER, 0.0], [2.0, EULER**3]],
                    base_point=[[EULER**3, 0.0], [4.0, EULER**4]],
                    expected=[[-2.0 * EULER**3, 0.0], [-2.0, -1 * EULER**4]],
                ),
                dict(
                    n=2,
                    point=[
                        [[gs.exp(-2.0), 0.0], [0.0, gs.exp(2.0)]],
                        [[gs.exp(-3.0), 0.0], [2.0, gs.exp(3.0)]],
                    ],
                    base_point=[[[1.0, 0.0], [-1.0, 1.0]], [[1.0, 0.0], [0.0, 1.0]]],
                    expected=[[[-2.0, 0.0], [1.0, 2.0]], [[-3.0, 0.0], [2.0, 3.0]]],
                ),
            ]
            return self.generate_tests(smoke_data)

        def squared_dist_data(self):
            smoke_data = [
                dict(
                    n=2,
                    point_a=[[EULER, 0.0], [2.0, EULER**3]],
                    point_b=[[EULER**3, 0.0], [4.0, EULER**4]],
                    expected=9,
                ),
                dict(
                    n=2,
                    point_a=[
                        [[EULER, 0.0], [2.0, EULER**3]],
                        [[EULER, 0.0], [4.0, EULER**3]],
                    ],
                    point_b=[
                        [[EULER**3, 0.0], [4.0, EULER**4]],
                        [[EULER**3, 0.0], [7.0, EULER**4]],
                    ],
                    expected=[9, 14],
                ),
            ]
            return self.generate_tests(smoke_data)

        def exp_shape_data(self):
            return self._exp_shape_data(
                self.metric_args_list,
                self.space_list,
                self.shape_list,
                self.batch_size_list,
            )

        def log_shape_data(self):
            return self._log_shape_data(
                self.metric_args_list,
                self.space_list,
                self.batch_size_list,
            )

        def squared_dist_is_symmetric_data(self):
            return self._squared_dist_is_symmetric_data(
                self.metric_args_list,
                self.space_list,
                self.n_points_a_list,
                self.n_points_b_list,
                atol=gs.atol * 1000,
            )

        def exp_belongs_data(self):
            return self._exp_belongs_data(
                self.metric_args_list,
                self.space_list,
                self.shape_list,
                self.n_samples_list,
                belongs_atol=gs.atol * 1000,
            )

        def log_is_tangent_data(self):
            return self._log_is_tangent_data(
                self.metric_args_list,
                self.space_list,
                self.n_samples_list,
                is_tangent_atol=gs.atol * 1000,
            )

        def geodesic_ivp_belongs_data(self):
            return self._geodesic_ivp_belongs_data(
                self.metric_args_list,
                self.space_list,
                self.shape_list,
                self.n_points_list,
                belongs_atol=gs.atol * 1000,
            )

        def geodesic_bvp_belongs_data(self):
            return self._geodesic_bvp_belongs_data(
                self.metric_args_list,
                self.space_list,
                self.n_points_list,
                belongs_atol=gs.atol * 1000,
            )

        def log_exp_composition_data(self):
            return self._log_exp_composition_data(
                self.metric_args_list,
                self.space_list,
                self.n_samples_list,
                rtol=gs.rtol * 100,
                atol=gs.atol * 10000,
            )

        def exp_log_composition_data(self):
            return self._exp_log_composition_data(
                self.metric_args_list,
                self.space_list,
                self.shape_list,
                self.n_samples_list,
                rtol=gs.rtol * 100,
                atol=gs.atol * 10000,
            )

        def exp_ladder_parallel_transport_data(self):
            return self._exp_ladder_parallel_transport_data(
                self.metric_args_list,
                self.space_list,
                self.shape_list,
                self.n_samples_list,
                self.n_rungs_list,
                self.alpha_list,
                self.scheme_list,
            )

        def exp_geodesic_ivp_data(self):
            return self._exp_geodesic_ivp_data(
                self.metric_args_list,
                self.space_list,
                self.shape_list,
                self.n_samples_list,
                self.n_points_list,
                rtol=gs.rtol * 100000,
                atol=gs.atol * 100000,
            )

        def parallel_transport_ivp_is_isometry_data(self):
            return self._parallel_transport_ivp_is_isometry_data(
                self.metric_args_list,
                self.space_list,
                self.shape_list,
                self.n_samples_list,
                is_tangent_atol=gs.atol * 1000,
                atol=gs.atol * 1000,
            )

        def parallel_transport_bvp_is_isometry_data(self):
            return self._parallel_transport_bvp_is_isometry_data(
                self.metric_args_list,
                self.space_list,
                self.shape_list,
                self.n_samples_list,
                is_tangent_atol=gs.atol * 1000,
                atol=gs.atol * 1000,
            )

    testing_data = TestDataCholeskyMetric()

    def test_diag_inner_product(
        self, n, tangent_vec_a, tangent_vec_b, base_point, expected
    ):
        result = self.metric(n).diag_inner_product(
            gs.array(tangent_vec_a), gs.array(tangent_vec_b), gs.array(base_point)
        )
        self.assertAllClose(result, gs.array(expected))

    def test_strictly_lower_inner_product(
        self, n, tangent_vec_a, tangent_vec_b, expected
    ):
        result = self.metric(n).strictly_lower_inner_product(
            gs.array(tangent_vec_a), gs.array(tangent_vec_b)
        )
        self.assertAllClose(result, gs.array(expected))

    def test_inner_product(self, n, tangent_vec_a, tangent_vec_b, base_point, expected):
        result = self.metric(n).inner_product(
            gs.array(tangent_vec_a), gs.array(tangent_vec_b), gs.array(base_point)
        )
        self.assertAllClose(result, gs.array(expected))

    def test_exp(self, n, tangent_vec, base_point, expected):
        result = self.metric(n).exp(gs.array(tangent_vec), gs.array(base_point))
        self.assertAllClose(result, gs.array(expected))

    def test_log(self, n, point, base_point, expected):
        result = self.metric(n).log(gs.array(point), gs.array(base_point))
        self.assertAllClose(result, gs.array(expected))

    def test_squared_dist(self, n, point_a, point_b, expected):
        result = self.metric(n).squared_dist(gs.array(point_a), gs.array(point_b))
        self.assertAllClose(result, gs.array(expected))



"""Unit tests for the General Linear group."""

import random

import geomstats.backend as gs
from geomstats.geometry.general_linear import GeneralLinear
from tests.conftest import TestCase
from tests.data_generation import LieGroupTestData
from tests.parametrizers import LieGroupParametrizer


class TestGeneralLinear(TestCase, metaclass=LieGroupParametrizer):
    space = group = GeneralLinear
    skip_test_to_tangent_is_tangent = True
    skip_test_exp_log_composition = True
    skip_test_log_exp_composition = True

    class TestDataGeneralLinear(LieGroupTestData):
        n_list = random.sample(range(2, 5), 2)
        positive_det_list = [True, False]
        space_args_list = list(zip(n_list, positive_det_list))
        shape_list = [(n, n) for n in n_list]
        n_samples_list = random.sample(range(2, 5), 2)
        n_points_list = random.sample(range(2, 5), 2)
        n_vecs_list = random.sample(range(2, 5), 2)

        def belongs_data(self):
            smoke_data = [
                dict(n=3, mat=gs.eye(3), expected=True),
                dict(n=3, mat=gs.ones((3, 3)), expected=False),
                dict(n=3, mat=gs.ones(3), expected=False),
                dict(n=3, mat=[gs.eye(3), gs.ones((3, 3))], expected=[True, False]),
            ]
            return self.generate_tests(smoke_data)

        def compose_data(self):
            smoke_data = [
                dict(
                    n=2,
                    mat1=[[1.0, 0.0], [0.0, 2.0]],
                    mat2=[[2.0, 0.0], [0.0, 1.0]],
                    expected=2.0 * GeneralLinear(2).identity,
                )
            ]
            return self.generate_tests(smoke_data)

        def inv_data(self):
            mat_a = [[0.0, 1.0, 0.0], [1.0, 0.0, 0.0], [0.0, 0.0, 1.0]]
            smoke_data = [
                dict(
                    n=3,
                    mat=gs.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 10.0]]),
                    expected=(
                        1.0
                        / 3.0
                        * gs.array(
                            [[-2.0, -4.0, 3.0], [-2.0, 11.0, -6.0], [3.0, -6.0, 3.0]]
                        )
                    ),
                ),
                dict(
                    n=3,
                    mat=gs.array([mat_a, -gs.eye(3, 3)]),
                    expected=gs.array([mat_a, -gs.eye(3, 3)]),
                ),
            ]
            return self.generate_tests(smoke_data)

        def exp_data(self):
            smoke_data = [
                dict(
                    n=3,
                    tangent_vec=[
                        [[2.0, 0.0, 0.0], [0.0, 3.0, 0.0], [0.0, 0.0, 4.0]],
                        [[1.0, 0.0, 0.0], [0.0, 5.0, 0.0], [0.0, 0.0, 6.0]],
                    ],
                    base_point=None,
                    expected=[
                        [
                            [7.38905609, 0.0, 0.0],
                            [0.0, 20.0855369, 0.0],
                            [0.0, 0.0, 54.5981500],
                        ],
                        [
                            [2.718281828, 0.0, 0.0],
                            [0.0, 148.413159, 0.0],
                            [0.0, 0.0, 403.42879349],
                        ],
                    ],
                )
            ]
            return self.generate_tests(smoke_data)

        def log_data(self):
            smoke_data = [
                dict(
                    n=3,
                    tangent_vec=[
                        [[2.0, 0.0, 0.0], [0.0, 3.0, 0.0], [0.0, 0.0, 4.0]],
                        [[1.0, 0.0, 0.0], [0.0, 5.0, 0.0], [0.0, 0.0, 6.0]],
                    ],
                    base_point=None,
                    expected=[
                        [
                            [0.693147180, 0.0, 0.0],
                            [0.0, 1.09861228866, 0.0],
                            [0.0, 0.0, 1.38629436],
                        ],
                        [
                            [0.0, 0.0, 0.0],
                            [0.0, 1.609437912, 0.0],
                            [0.0, 0.0, 1.79175946],
                        ],
                    ],
                )
            ]
            return self.generate_tests(smoke_data)

        def orbit_data(self):
            point = gs.array([[gs.exp(4.0), 0.0], [0.0, gs.exp(2.0)]])
            sqrt = gs.array([[gs.exp(2.0), 0.0], [0.0, gs.exp(1.0)]])
            identity = GeneralLinear(2).identity
            time = gs.linspace(0.0, 1.0, 3)
            smoke_data = [
                dict(
                    n=2,
                    point=point,
                    base_point=identity,
                    time=time,
                    expected=gs.array([identity, sqrt, point]),
                ),
                dict(
                    n=2,
                    point=[point, point],
                    base_point=identity,
                    time=time,
                    expected=[
                        gs.array([identity, sqrt, point]),
                        gs.array([identity, sqrt, point]),
                    ],
                ),
            ]
            return self.generate_tests(smoke_data)

        def random_point_belongs_data(self):
            smoke_space_args_list = [(2, True), (3, True), (2, False)]
            smoke_n_points_list = [1, 2, 1]
            return self._random_point_belongs_data(
                smoke_space_args_list,
                smoke_n_points_list,
                self.space_args_list,
                self.n_points_list,
            )

        def projection_belongs_data(self):
            return self._projection_belongs_data(
                self.space_args_list, self.shape_list, self.n_samples_list
            )

        def to_tangent_is_tangent_data(self):
            return self._to_tangent_is_tangent_data(
                GeneralLinear,
                self.space_args_list,
                self.shape_list,
                self.n_vecs_list,
            )

        def exp_log_composition_data(self):
            return self._exp_log_composition_data(
                GeneralLinear,
                self.space_args_list,
                self.shape_list,
                self.n_samples_list,
            )

        def log_exp_composition_data(self):
            return self._log_exp_composition_data(
                GeneralLinear, self.space_args_list, self.n_samples_list
            )

    testing_data = TestDataGeneralLinear()

    def test_belongs(self, n, point, expected):
        group = self.space(n)
        self.assertAllClose(group.belongs(gs.array(point)), gs.array(expected))

    def test_compose(self, n, mat1, mat2, expected):
        group = self.space(n)
        self.assertAllClose(
            group.compose(gs.array(mat1), gs.array(mat2)), gs.array(expected)
        )

    def test_inv(self, n, mat, expected):
        group = self.space(n)
        self.assertAllClose(group.inverse(gs.array(mat)), gs.array(expected))

    def test_exp(self, n, tangent_vec, base_point, expected):
        group = self.space(n)
        expected = gs.cast(gs.array(expected), gs.float64)
        tangent_vec = gs.cast(gs.array(tangent_vec), gs.float64)
        base_point = (
            None if base_point is None else gs.cast(gs.array(base_point), gs.float64)
        )
        self.assertAllClose(group.exp(tangent_vec, base_point), gs.array(expected))

    def test_log(self, n, point, base_point, expected):
        group = self.space(n)
        expected = gs.cast(gs.array(expected), gs.float64)
        point = gs.cast(gs.array(point), gs.float64)
        base_point = (
            None if base_point is None else gs.cast(gs.array(base_point), gs.float64)
        )
        self.assertAllClose(group.log(point, base_point), expected)

    def test_orbit(self, n, point, base_point, time, expected):
        group = self.space(n)
        result = group.orbit(gs.array(point), gs.array(base_point))(time)
        self.assertAllClose(result, gs.array(expected))



"""Unit tests for TimeSeriesCovariance class."""

import geomstats.backend as gs
import geomstats.tests
from geomstats.datasets.prepare_emg_data import TimeSeriesCovariance
from geomstats.datasets.utils import load_emg


@geomstats.tests.np_and_autograd_only
class TestPrepareEmgData(geomstats.tests.TestCase):
    """Class for testing the covariance creation from time series."""

    def setup_method(self):
        """Set up function."""
        self.n_steps = 100
        self.n_elec = 8
        self.label_map = {"rock": 0, "scissors": 1, "paper": 2, "ok": 3}
        self.margin = 1000
        data = load_emg()
        data = data[data.label != "rest"]
        self.emg_data = {
            "time_vec": gs.array(data.time),
            "raw_data": gs.array(data[["c{}".format(i) for i in range(8)]]),
            "label": gs.array(data.label),
            "exp": gs.array(data.exp),
        }

        self.cov_transformer = TimeSeriesCovariance(
            self.emg_data, self.n_steps, self.n_elec, self.label_map, self.margin
        )
        self.cov_transformer.transform()

    def test_covariance_shape(self):
        """Test the shape of the covariance matrices."""
        result_shape = (len(self.cov_transformer.batches), self.n_elec, self.n_elec)
        self.assertTrue(self.cov_transformer.covs.shape == result_shape)

    def test_covec_shape(self):
        """Test the shape of the vectorized covariance."""
        dim_vec = int(self.n_elec * (self.n_elec + 1) / 2)
        result_shape = (len(self.cov_transformer.batches), dim_vec)
        shape = self.cov_transformer.covecs.shape
        self.assertTrue(shape == result_shape)

    def test_diag_shape(self):
        """Test the shape of the diagonal."""
        result_shape = (len(self.cov_transformer.batches), self.n_elec)
        shape = self.cov_transformer.diags.shape
        self.assertTrue(shape == result_shape)



"""Unit tests for special euclidean group SE(n).

Note: Only the *canonical* left- and right- invariant
metrics on SE(3) are tested here. Other invariant
metrics are tested with the tests of the invariant_metric
module.
"""

import warnings

import geomstats.backend as gs
import geomstats.tests
import tests.helper as helper
from geomstats.geometry.invariant_metric import InvariantMetric
from geomstats.geometry.special_euclidean import SpecialEuclidean

# Tolerance for errors on predicted vectors, relative to the *norm*
# of the vector, as opposed to the standard behavior of gs.allclose
# where it is relative to each element of the array

ATOL = 1e-5


class TestSpecialEuclidean3Methods(geomstats.tests.TestCase):
    def setup_method(self):
        warnings.simplefilter("ignore", category=ImportWarning)
        gs.random.seed(1234)

        group = SpecialEuclidean(n=3, point_type="vector")

        # Points

        # -- Rotation vectors with angles
        # 0, close to 0, closely lower than pi, pi,
        # between pi and 2pi, closely larger than 2pi, 2pi,
        # and closely larger than 2pi
        with_angle_0 = gs.zeros(6)
        with_angle_close_0 = 1e-10 * gs.array(
            [1.0, -1.0, 1.0, 0.0, 0.0, 0.0]
        ) + gs.array([0.0, 0.0, 0.0, 1.0, 5.0, 2])
        with_angle_close_pi_low = (gs.pi - 1e-9) / gs.sqrt(2.0) * gs.array(
            [0.0, 1.0, -1.0, 0.0, 0.0, 0.0]
        ) + gs.array([0.0, 0.0, 0.0, -100.0, 0.0, 2.0])
        with_angle_pi = gs.pi / gs.sqrt(3.0) * gs.array(
            [1.0, 1.0, -1.0, 0.0, 0.0, 0.0]
        ) + gs.array([0.0, 0.0, 0.0, -10.2, 0.0, 2.6])
        with_angle_close_pi_high = (gs.pi + 1e-9) / gs.sqrt(3.0) * gs.array(
            [-1.0, 1.0, -1.0, 0.0, 0.0, 0.0]
        ) + gs.array([0.0, 0.0, 0.0, -100.0, 0.0, 2.0])
        with_angle_in_pi_2pi = (gs.pi + 0.3) / gs.sqrt(5.0) * gs.array(
            [-2.0, 1.0, 0.0, 0.0, 0.0, 0.0]
        ) + gs.array([0.0, 0.0, 0.0, -100.0, 0.0, 2.0])
        with_angle_close_2pi_low = (2 * gs.pi - 1e-9) / gs.sqrt(6.0) * gs.array(
            [2.0, 1.0, -1.0, 0.0, 0.0, 0.0]
        ) + gs.array([0.0, 0.0, 0.0, 8.0, 555.0, -2.0])
        with_angle_2pi = 2.0 * gs.pi / gs.sqrt(3.0) * gs.array(
            [1.0, 1.0, -1.0, 0.0, 0.0, 0.0]
        ) + gs.array([0.0, 0.0, 0.0, 1.0, 8.0, -10.0])
        with_angle_close_2pi_high = (2.0 * gs.pi + 1e-9) / gs.sqrt(2.0) * gs.array(
            [1.0, 0.0, -1.0, 0.0, 0.0, 0.0]
        ) + gs.array([0.0, 0.0, 0.0, 1.0, 8.0, -10.0])

        point_1 = gs.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6])
        point_2 = gs.array([0.5, 0.0, -0.3, 0.4, 5.0, 60.0])

        translation_large = gs.array([0.0, 0.0, 0.0, 0.4, 0.5, 0.6])
        translation_small = gs.array([0.0, 0.0, 0.0, 0.5, 0.6, 0.7])
        rot_with_parallel_trans = gs.array([gs.pi / 3.0, 0.0, 0.0, 1.0, 0.0, 0.0])

        elements_all = {
            "with_angle_0": with_angle_0,
            "with_angle_close_0": with_angle_close_0,
            "with_angle_close_pi_low": with_angle_close_pi_low,
            "with_angle_pi": with_angle_pi,
            "with_angle_close_pi_high": with_angle_close_pi_high,
            "with_angle_in_pi_2pi": with_angle_in_pi_2pi,
            "with_angle_close_2pi_low": with_angle_close_2pi_low,
            "with_angle_2pi": with_angle_2pi,
            "with_angle_close_2pi_high": with_angle_close_2pi_high,
            "translation_large": translation_large,
            "translation_small": translation_small,
            "point_1": point_1,
            "point_2": point_2,
            "rot_with_parallel_trans": rot_with_parallel_trans,
        }
        elements = elements_all
        if geomstats.tests.tf_backend():
            # Tf is extremely slow
            elements = {"point_1": point_1, "point_2": point_2}

        elements_matrices_all = {
            key: group.matrix_from_vector(elements_all[key]) for key in elements_all
        }
        elements_matrices = elements_matrices_all

        # Metrics - only diagonals
        diag_mat_at_identity = gs.eye(6) * gs.array([2.0, 2.0, 2.0, 3.0, 3.0, 3.0])

        left_diag_metric = InvariantMetric(
            group=group,
            metric_mat_at_identity=diag_mat_at_identity,
            left_or_right="left",
        )
        right_diag_metric = InvariantMetric(
            group=group,
            metric_mat_at_identity=diag_mat_at_identity,
            left_or_right="right",
        )

        # mat_at_identity = 7 * gs.eye(group.dim)

        # left_metric = InvariantMetric(
        #            group=group,
        #            metric_mat_at_identity=mat_at_identity,
        #            left_or_right='left')
        # right_metric = InvariantMetric(
        #            group=group,
        #            metric_mat_at_identity=mat_at_identity,
        #            left_or_right='right')

        metrics_all = {
            "left_canonical": group.left_canonical_metric,
            "right_canonical": group.right_canonical_metric,
            "left_diag": left_diag_metric,
            "right_diag": right_diag_metric,
        }
        # FIXME:
        # 'left': left_metric,
        # 'right': right_metric}
        metrics = metrics_all
        if geomstats.tests.tf_backend():
            metrics = {"left_diag": left_diag_metric}

        self.group = group
        self.metrics_all = metrics_all
        self.metrics = metrics
        self.elements_all = elements_all
        self.elements = elements
        self.elements_matrices_all = elements_matrices_all
        self.elements_matrices = elements_matrices
        self.angles_close_to_pi_all = [
            "with_angle_close_pi_low",
            "with_angle_pi",
            "with_angle_close_pi_high",
        ]
        self.angles_close_to_pi = self.angles_close_to_pi_all
        if geomstats.tests.tf_backend():
            self.angles_close_to_pi = ["with_angle_close_pi_low"]

        self.n_samples = 4

    def test_random_and_belongs(self):
        """Checks random_point and belongs

        Test that the random uniform method samples
        on the special euclidean group.
        """
        base_point = self.group.random_point()
        result = self.group.belongs(base_point)
        expected = True
        self.assertAllClose(result, expected)

    def test_random_and_belongs_vectorization(self):
        n_samples = self.n_samples
        points = self.group.random_point(n_samples=n_samples)
        result = self.group.belongs(points)
        expected = gs.array([True] * n_samples)
        self.assertAllClose(result, expected)

    def test_regularize(self):
        point = gs.random.rand(6) * 2 * gs.pi
        regularized = self.group.regularize(point)
        result = 0 <= gs.linalg.norm(regularized[:3]) <= gs.pi
        self.assertTrue(result)

    @geomstats.tests.np_autograd_and_tf_only
    def test_regularize_extreme_cases(self):
        point = self.elements_all["with_angle_0"]
        result = self.group.regularize(point)
        expected = point
        self.assertAllClose(result, expected)

        less_than_pi = ["with_angle_close_0", "with_angle_close_pi_low"]
        for angle_type in less_than_pi:
            point = self.elements_all[angle_type]
            result = self.group.regularize(point)
            expected = point
            self.assertAllClose(result, expected)

        if not geomstats.tests.tf_backend():
            angle_type = "with_angle_pi"
            point = self.elements_all[angle_type]
            result = self.group.regularize(point)

            expected = point

            self.assertAllClose(result, expected)

            angle_type = "with_angle_close_pi_high"
            point = self.elements_all[angle_type]
            result = self.group.regularize(point)
            norm = gs.linalg.norm(point[:3])
            expected_rot = gs.concatenate(
                [point[:3] / norm * (norm - 2 * gs.pi), gs.zeros(3)], axis=0
            )
            expected_trans = gs.concatenate([gs.zeros(3), point[3:6]], axis=0)
            expected = expected_rot + expected_trans
            self.assertAllClose(result, expected)

            in_pi_2pi = ["with_angle_in_pi_2pi", "with_angle_close_2pi_low"]

            for angle_type in in_pi_2pi:
                point = self.elements_all[angle_type]
                angle = gs.linalg.norm(point[:3])
                new_angle = gs.pi - (angle - gs.pi)

                result = self.group.regularize(point)
                expected_rot = gs.concatenate(
                    [-new_angle * (point[:3] / angle), gs.zeros(3)], axis=0
                )
                expected_trans = gs.concatenate([gs.zeros(3), point[3:6]], axis=0)
                expected = expected_rot + expected_trans

                self.assertAllClose(result, expected)

            angle_type = "with_angle_2pi"
            point = self.elements_all[angle_type]
            result = self.group.regularize(point)
            expected = gs.concatenate([gs.zeros(3), point[3:6]], axis=0)
            self.assertAllClose(result, expected)

            angle_type = "with_angle_close_2pi_high"
            point = self.elements_all[angle_type]
            angle = gs.linalg.norm(point[:3])
            new_angle = angle - 2 * gs.pi

            result = self.group.regularize(point)

            expected_rot = gs.concatenate(
                [new_angle * point[:3] / angle, gs.zeros(3)], axis=0
            )
            expected_trans = gs.concatenate([gs.zeros(3), point[3:6]], axis=0)
            expected = expected_rot + expected_trans
            self.assertAllClose(result, expected)

    def test_regularize_vectorization(self):
        n_samples = self.n_samples
        points = self.group.random_point(n_samples=n_samples)
        regularized_points = self.group.regularize(points)

        self.assertAllClose(
            gs.shape(regularized_points),
            (n_samples, *self.group.get_point_type_shape()),
        )
        result = 0 <= gs.linalg.norm(regularized_points[:3]) <= gs.pi
        self.assertTrue(result)

    def test_compose(self):
        # Composition by identity, on the right
        # Expect the original transformation
        point = self.elements_all["point_1"]
        result = self.group.compose(point, self.group.identity)
        expected = point
        self.assertAllClose(result, expected)

        if not geomstats.tests.tf_backend():
            # Composition by identity, on the left
            # Expect the original transformation
            result = self.group.compose(self.group.identity, point)
            expected = point
            self.assertAllClose(result, expected)

            # Composition of translations (no rotational part)
            # Expect the sum of the translations
            result = self.group.compose(
                self.elements_all["translation_small"],
                self.elements_all["translation_large"],
            )
            expected = (
                self.elements_all["translation_small"]
                + self.elements_all["translation_large"]
            )
            self.assertAllClose(result, expected)

    def test_compose_and_inverse(self):
        point = self.elements_all["point_1"]
        inv_point = self.group.inverse(point)
        # Compose transformation by its inverse on the right
        # Expect the group identity
        result = self.group.compose(point, inv_point)
        expected = self.group.identity
        self.assertAllClose(result, expected)

        if not geomstats.tests.tf_backend():
            # Compose transformation by its inverse on the left
            # Expect the group identity
            result = self.group.compose(inv_point, point)
            expected = self.group.identity
            self.assertAllClose(result, expected)

    def test_compose_vectorization(self):
        n_samples = self.n_samples
        n_points_a = self.group.random_point(n_samples=n_samples)
        n_points_b = self.group.random_point(n_samples=n_samples)
        one_point = self.group.random_point(n_samples=1)

        result = self.group.compose(one_point, n_points_a)
        self.assertAllClose(
            gs.shape(result), (n_samples, *self.group.get_point_type_shape())
        )

        result = self.group.compose(n_points_a, one_point)

        if not geomstats.tests.tf_backend():
            self.assertAllClose(
                gs.shape(result), (n_samples, *self.group.get_point_type_shape())
            )

            result = self.group.compose(n_points_a, n_points_b)
            self.assertAllClose(
                gs.shape(result), (n_samples, *self.group.get_point_type_shape())
            )

    def test_inverse_vectorization(self):
        n_samples = self.n_samples
        points = self.group.random_point(n_samples=n_samples)
        result = self.group.inverse(points)
        self.assertAllClose(
            gs.shape(result), (n_samples, *self.group.get_point_type_shape())
        )

    def test_left_jacobian_vectorization(self):
        n_samples = self.n_samples
        points = self.group.random_point(n_samples=n_samples)
        result = self.group.jacobian_translation(point=points, left_or_right="left")
        self.assertAllClose(
            gs.shape(result),
            (
                n_samples,
                *self.group.get_point_type_shape(),
                *self.group.get_point_type_shape(),
            ),
        )

    def test_exp_from_identity_vectorization(self):
        n_samples = self.n_samples
        for metric in self.metrics.values():
            tangent_vecs = self.group.random_point(n_samples=n_samples)
            result = metric.exp_from_identity(tangent_vecs)

            self.assertAllClose(
                gs.shape(result), (n_samples, *self.group.get_point_type_shape())
            )

            if geomstats.tests.tf_backend():
                break

    def test_log_from_identity_vectorization(self):
        n_samples = self.n_samples
        for metric in self.metrics.values():
            points = self.group.random_point(n_samples=n_samples)
            result = metric.log_from_identity(points)

            self.assertAllClose(
                gs.shape(result), (n_samples, *self.group.get_point_type_shape())
            )

            if geomstats.tests.tf_backend():
                break

    def test_exp_vectorization(self):
        n_samples = self.n_samples

        for metric_type in self.metrics:
            metric = self.metrics[metric_type]

            one_tangent_vec = self.group.random_point(n_samples=1)
            one_base_point = self.group.random_point(n_samples=1)
            n_tangent_vec = self.group.random_point(n_samples=n_samples)
            n_base_point = self.group.random_point(n_samples=n_samples)

            # Test with the 1 base point, and n tangent vecs
            result = metric.exp(n_tangent_vec, one_base_point)
            self.assertAllClose(
                gs.shape(result), (n_samples, *self.group.get_point_type_shape())
            )

            if geomstats.tests.tf_backend():
                break

            # Test with the several base point, and one tangent vec
            result = metric.exp(one_tangent_vec, n_base_point)
            self.assertAllClose(
                gs.shape(result), (n_samples, *self.group.get_point_type_shape())
            )

            # Test with the same number n of base point and n tangent vec
            result = metric.exp(n_tangent_vec, n_base_point)
            self.assertAllClose(
                gs.shape(result), (n_samples, *self.group.get_point_type_shape())
            )

    def test_log_vectorization(self):
        n_samples = self.n_samples
        for metric_type in self.metrics:
            metric = self.metrics[metric_type]

            one_point = self.group.random_point(n_samples=1)
            one_base_point = self.group.random_point(n_samples=1)
            n_point = self.group.random_point(n_samples=n_samples)
            n_base_point = self.group.random_point(n_samples=n_samples)

            # Test with the 1 base point, and several different points
            result = metric.log(n_point, one_base_point)
            self.assertAllClose(
                gs.shape(result), (n_samples, *self.group.get_point_type_shape())
            )

            if geomstats.tests.tf_backend():
                break

            # Test with the several base point, and 1 point
            result = metric.log(one_point, n_base_point)
            self.assertAllClose(
                gs.shape(result), (n_samples, *self.group.get_point_type_shape())
            )

            # Test with the same number n of base point and point
            result = metric.log(n_point, n_base_point)
            self.assertAllClose(
                gs.shape(result), (n_samples, *self.group.get_point_type_shape())
            )

    @geomstats.tests.np_and_autograd_only
    def test_group_exp_from_identity_vectorization(self):
        n_samples = self.n_samples
        tangent_vecs = self.group.random_point(n_samples=n_samples)
        result = self.group.exp_from_identity(tangent_vecs)

        self.assertAllClose(
            gs.shape(result), (n_samples, *self.group.get_point_type_shape())
        )

    @geomstats.tests.np_and_autograd_only
    def test_group_log_from_identity_vectorization(self):
        n_samples = self.n_samples
        points = self.group.random_point(n_samples=n_samples)
        result = self.group.log_from_identity(points)

        self.assertAllClose(
            gs.shape(result), (n_samples, *self.group.get_point_type_shape())
        )

    @geomstats.tests.np_and_autograd_only
    def test_group_exp_vectorization(self):
        n_samples = self.n_samples
        # Test with the 1 base_point, and several different tangent_vecs
        tangent_vecs = self.group.random_point(n_samples=n_samples)
        base_point = self.group.random_point(n_samples=1)
        result = self.group.exp(tangent_vecs, base_point)

        self.assertAllClose(
            gs.shape(result), (n_samples, *self.group.get_point_type_shape())
        )

        if not geomstats.tests.tf_backend():
            # Test with the same number of base_points and tangent_vecs
            tangent_vecs = self.group.random_point(n_samples=n_samples)
            base_points = self.group.random_point(n_samples=n_samples)
            result = self.group.exp(tangent_vecs, base_points)

            self.assertAllClose(
                gs.shape(result), (n_samples, *self.group.get_point_type_shape())
            )

            # Test with the several base_points, and 1 tangent_vec
            tangent_vec = self.group.random_point(n_samples=1)
            base_points = self.group.random_point(n_samples=n_samples)
            result = self.group.exp(tangent_vec, base_points)

            self.assertAllClose(
                gs.shape(result), (n_samples, *self.group.get_point_type_shape())
            )

    @geomstats.tests.np_and_autograd_only
    def test_group_log_vectorization(self):
        n_samples = self.n_samples
        # Test with the 1 base point, and several different points
        points = self.group.random_point(n_samples=n_samples)
        base_point = self.group.random_point(n_samples=1)
        result = self.group.log(points, base_point)

        self.assertAllClose(
            gs.shape(result), (n_samples, *self.group.get_point_type_shape())
        )

        if not geomstats.tests.tf_backend():

            # Test with the same number of base points and points
            points = self.group.random_point(n_samples=n_samples)
            base_points = self.group.random_point(n_samples=n_samples)
            result = self.group.log(points, base_points)

            self.assertAllClose(
                gs.shape(result), (n_samples, *self.group.get_point_type_shape())
            )

            # Test with the several base points, and 1 point
            point = self.group.random_point(n_samples=1)
            base_points = self.group.random_point(n_samples=n_samples)
            result = self.group.log(point, base_points)

            self.assertAllClose(
                gs.shape(result), (n_samples, *self.group.get_point_type_shape())
            )

    @geomstats.tests.np_and_autograd_only
    def test_group_exp_from_identity(self):
        # Group exponential of a translation (no rotational part)
        # Expect the original translation
        tangent_vec = self.elements_all["translation_small"]
        result = self.group.exp(base_point=self.group.identity, tangent_vec=tangent_vec)
        expected = tangent_vec
        self.assertAllClose(result, expected)

        if not geomstats.tests.tf_backend():
            # Group exponential of a transformation
            # where translation is parallel to rotation axis
            # Expect the original transformation
            tangent_vec = self.elements_all["rot_with_parallel_trans"]
            result = self.group.exp(
                base_point=self.group.identity, tangent_vec=tangent_vec
            )
            expected = tangent_vec
            self.assertAllClose(result, expected)

    @geomstats.tests.np_and_autograd_only
    def test_group_log_from_identity(self):
        # Group logarithm of a translation (no rotational part)
        # Expect the original translation
        point = self.elements_all["translation_small"]
        result = self.group.log(base_point=self.group.identity, point=point)
        expected = point
        self.assertAllClose(result, expected)

        if not geomstats.tests.tf_backend():
            # Group logarithm of a transformation
            # where translation is parallel to rotation axis
            # Expect the original transformation
            point = self.elements_all["rot_with_parallel_trans"]
            result = self.group.log(base_point=self.group.identity, point=point)
            expected = point
            self.assertAllClose(result, expected)

    @geomstats.tests.np_and_autograd_only
    def test_group_log_then_exp_from_identity(self):
        """
        Test that the group exponential from the identity
        and the group logarithm from the identity
        are inverse.
        Expect their composition to give the identity function.
        """
        for element_type in self.elements:
            if element_type in self.angles_close_to_pi:
                continue
            point = self.elements[element_type]
            result = helper.group_log_then_exp_from_identity(
                group=self.group, point=point
            )
            expected = self.group.regularize(point)
            self.assertAllClose(result, expected)

            if geomstats.tests.tf_backend():
                break

    @geomstats.tests.np_and_autograd_only
    def test_group_log_then_exp_from_identity_with_angles_close_to_pi(self):
        """
        Test that the group exponential from the identity
        and the group logarithm from the identity
        are inverse.
        Expect their composition to give the identity function.
        """
        angle_types = self.angles_close_to_pi
        for element_type in angle_types:
            point = self.elements_all[element_type]
            result = helper.group_log_then_exp_from_identity(
                group=self.group, point=point
            )
            expected = self.group.regularize(point)

            inv_expected = gs.concatenate([-expected[:3], expected[3:6]])

            self.assertTrue(
                gs.allclose(result, expected) or gs.allclose(result, inv_expected)
            )

            if geomstats.tests.tf_backend():
                break

    @geomstats.tests.np_and_autograd_only
    def test_group_exp(self):
        # Reference point is a translation (no rotational part)
        # so that the jacobian of the left-translation of the Lie group
        # is the 6x6 identity matrix
        # Tangent vector is a translation (no infinitesimal rotational part)
        # Expect the sum of the translation
        # with the translation of the reference point
        result = self.group.exp(
            base_point=self.elements_all["translation_small"],
            tangent_vec=self.elements_all["translation_large"],
        )
        expected = (
            self.elements_all["translation_small"]
            + self.elements_all["translation_large"]
        )
        self.assertAllClose(result, expected)

    @geomstats.tests.np_and_autograd_only
    def test_group_log(self):
        # Reference point is a translation (no rotational part)
        # so that the jacobian of the left-translation of the Lie group
        # is the 6x6 identity matrix
        # Point is a translation (no rotational part)
        # Expect the difference of the translation
        # by the translation of the reference point
        result = self.group.log(
            base_point=self.elements_all["translation_small"],
            point=self.elements_all["translation_large"],
        )
        expected = (
            self.elements_all["translation_large"]
            - self.elements_all["translation_small"]
        )

        self.assertAllClose(result, expected)

    @geomstats.tests.np_and_autograd_only
    def test_group_log_then_exp(self):
        """
        Test that the group exponential
        and the group logarithm are inverse.
        Expect their composition to give the identity function.
        """
        for base_point in self.elements.values():
            for element_type in self.elements:
                if element_type in self.angles_close_to_pi:
                    continue
                point = self.elements[element_type]

                result = helper.group_log_then_exp(
                    group=self.group, point=point, base_point=base_point
                )
                expected = self.group.regularize(point)
                self.assertAllClose(result, expected, atol=1e-4)

                if geomstats.tests.tf_backend():
                    break

    @geomstats.tests.np_and_autograd_only
    def test_group_exp_then_log(self):
        """
        Test that the group exponential
        and the group logarithm are inverse.
        Expect their composition to give the identity function.
        """
        for base_point_type in self.elements:
            base_point = self.elements[base_point_type]
            for element_type in self.elements:
                if element_type in self.angles_close_to_pi:
                    continue
                tangent_vec = self.elements[element_type]
                result = helper.group_exp_then_log(
                    group=self.group, tangent_vec=tangent_vec, base_point=base_point
                )
                metric = self.metrics_all["left_canonical"]
                expected = self.group.regularize_tangent_vec(
                    tangent_vec=tangent_vec, base_point=base_point, metric=metric
                )
                self.assertAllClose(result, expected, atol=1e-4)

                if geomstats.tests.tf_backend():
                    break

    @geomstats.tests.np_and_autograd_only
    def test_exp_from_identity_left(self):
        # Riemannian left-invariant metric given by
        # the canonical inner product on the lie algebra
        # Expect the identity function
        # because we use the Riemannian left logarithm with canonical
        # inner product to parameterize the transformations
        metric = self.metrics_all["left_canonical"]
        # General case
        tangent_rot_vec = gs.array([1.0, 1.0, 1.0])  # NB: Regularized
        tangent_translation = gs.array([1.0, 0.0, -3.0])
        tangent_vec = gs.concatenate([tangent_rot_vec, tangent_translation], axis=0)
        result = metric.exp_from_identity(tangent_vec)
        expected = tangent_vec

        self.assertAllClose(result, expected)

    @geomstats.tests.np_and_autograd_only
    def test_log_from_identity_left(self):
        # Riemannian left-invariant metric given by
        # the canonical inner product on the lie algebra
        # Expect the identity function
        # because we use the Riemannian left logarithm with canonical
        # inner product to parameterize the transformations

        metric = self.metrics_all["left_canonical"]
        # General case
        rot_vec = gs.array([0.1, 1, 0.9])  # NB: Regularized
        translation = gs.array([1.0, -19.0, -3.0])
        transfo = gs.concatenate([rot_vec, translation], axis=0)

        expected = transfo
        result = metric.log_from_identity(transfo)

        self.assertAllClose(result, expected)

        if not geomstats.tests.tf_backend():
            # Edge case: angle < epsilon, where angle = norm(rot_vec)
            rot_vec = gs.array([1e-8, 0.0, 1e-9])  # NB: Regularized
            translation = gs.array([10000.0, -5.9, -93])
            transfo = gs.concatenate([rot_vec, translation], axis=0)

            expected = transfo
            result = metric.log_from_identity(transfo)

            self.assertAllClose(result, expected)

    @geomstats.tests.np_and_autograd_only
    def test_exp_then_log_from_identity_left(self):
        """
        Test that the Riemannian left exponential from the identity
        and the Riemannian left logarithm from the identity
        are inverse.
        Expect their composition to give the identity function.
        """
        # Canonical inner product on the lie algebra

        for metric in [
            self.metrics_all["left_canonical"],
            self.metrics_all["left_diag"],
        ]:
            for angle_type in self.elements:
                if angle_type in self.angles_close_to_pi:
                    continue
                tangent_vec = self.elements[angle_type]
                result = helper.exp_then_log_from_identity(
                    metric=metric, tangent_vec=tangent_vec
                )
                expected = self.group.regularize_tangent_vec_at_identity(
                    tangent_vec, metric=metric
                )
                self.assertAllClose(result, expected)

                if geomstats.tests.tf_backend():
                    break

    @geomstats.tests.np_and_autograd_only
    def test_exp_then_log_from_identity_left_with_angles_close_to_pi(self):
        """
        Test that the Riemannian left exponential from the identity
        and the Riemannian left logarithm from the identity
        are inverse.
        Expect their composition to give the identity function.
        """
        angle_types = self.angles_close_to_pi
        # Canonical inner product on the lie algebra
        for metric in [
            self.metrics_all["left_canonical"],
            self.metrics_all["left_diag"],
        ]:
            for angle_type in angle_types:
                tangent_vec = self.elements_all[angle_type]
                result = helper.exp_then_log_from_identity(
                    metric=metric, tangent_vec=tangent_vec
                )
                expected = self.group.regularize_tangent_vec_at_identity(
                    tangent_vec=tangent_vec, metric=metric
                )
                inv_expected = gs.concatenate([-expected[:3], expected[3:6]])

                self.assertTrue(
                    gs.allclose(result, expected) or gs.allclose(result, inv_expected)
                )

                if geomstats.tests.tf_backend():
                    break

    @geomstats.tests.np_and_autograd_only
    def test_exp_then_log_from_identity_right(self):
        """
        Test that the Riemannian right exponential from the identity
        and the Riemannian right logarithm from the identity
        are inverse.
        Expect their composition to give the identity function.
        """
        # Canonical inner product on the lie algebra
        for metric in [
            self.metrics_all["right_canonical"],
            self.metrics_all["right_diag"],
        ]:
            for angle_type in self.elements:
                if angle_type in self.angles_close_to_pi:
                    continue
                tangent_vec = self.elements[angle_type]
                result = helper.exp_then_log_from_identity(
                    metric=metric, tangent_vec=tangent_vec
                )
                expected = self.group.regularize_tangent_vec_at_identity(
                    tangent_vec=tangent_vec, metric=metric
                )

                self.assertAllClose(result, expected)

                if geomstats.tests.tf_backend():
                    break

    @geomstats.tests.np_and_autograd_only
    def test_exp_then_log_from_identity_right_with_angles_close_to_pi(self):
        """
        Test that the Riemannian right exponential from the identity
        and the Riemannian right logarithm from the identity
        are inverse.
        Expect their composition to give the identity function.
        """
        angle_types = self.angles_close_to_pi
        # Canonical inner product on the lie algebra
        for metric in [
            self.metrics_all["right_canonical"],
            self.metrics_all["right_diag"],
        ]:
            for angle_type in angle_types:
                tangent_vec = self.elements_all[angle_type]
                result = helper.exp_then_log_from_identity(
                    metric=metric, tangent_vec=tangent_vec
                )
                expected = self.group.regularize_tangent_vec_at_identity(
                    tangent_vec=tangent_vec, metric=metric
                )
                inv_expected = gs.concatenate([-expected[:3], expected[3:6]])
                self.assertTrue(
                    gs.allclose(result, expected) or gs.allclose(result, inv_expected)
                )

    @geomstats.tests.np_and_autograd_only
    def test_exp_left(self):
        # Reference point is a translation (no rotational part)
        # so that the jacobian of the left-translation of the Lie group
        # is the 6x6 identity matrix
        metric = self.metrics_all["left_canonical"]
        rot_vec_base_point = gs.array([0.0, 0.0, 0.0])
        translation_base_point = gs.array([4.0, -1.0, 10000.0])
        transfo_base_point = gs.concatenate(
            [rot_vec_base_point, translation_base_point], axis=0
        )

        # Tangent vector is a translation (no infinitesimal rotational part)
        # Expect the sum of the translation
        # with the translation of the reference point
        rot_vec = gs.array([0.0, 0.0, 0.0])
        translation = gs.array([1.0, 0.0, -3.0])
        tangent_vec = gs.concatenate([rot_vec, translation], axis=0)

        result = metric.exp(base_point=transfo_base_point, tangent_vec=tangent_vec)
        expected = gs.concatenate(
            [gs.array([0.0, 0.0, 0.0]), gs.array([5.0, -1.0, 9997.0])], axis=0
        )
        self.assertAllClose(result, expected)

    @geomstats.tests.np_and_autograd_only
    def test_log_left(self):
        # Reference point is a translation (no rotational part)
        # so that the jacobian of the left-translation of the Lie group
        # is the 6x6 identity matrix
        metric = self.metrics_all["left_canonical"]
        rot_vec_base_point = gs.array([0.0, 0.0, 0.0])
        translation_base_point = gs.array([4.0, 0.0, 0.0])
        transfo_base_point = gs.concatenate(
            [rot_vec_base_point, translation_base_point], axis=0
        )

        # Point is a translation (no rotational part)
        # Expect the difference of the translation
        # by the translation of the reference point
        rot_vec = gs.array([0.0, 0.0, 0.0])
        translation = gs.array([-1.0, -1.0, -1.2])
        point = gs.concatenate([rot_vec, translation], axis=0)

        expected = gs.concatenate(
            [gs.array([0.0, 0.0, 0.0]), gs.array([-5.0, -1.0, -1.2])], axis=0
        )

        result = metric.log(base_point=transfo_base_point, point=point)

        self.assertAllClose(result, expected)

    @geomstats.tests.np_and_autograd_only
    def test_log_then_exp_left(self):
        """
        Test that the Riemannian left exponential and the
        Riemannian left logarithm are inverse.
        Expect their composition to give the identity function.
        """
        for metric in [
            self.metrics_all["left_canonical"],
            self.metrics_all["left_diag"],
        ]:
            for base_point_type in self.elements:
                base_point = self.elements[base_point_type]
                for element_type in self.elements:
                    if element_type in self.angles_close_to_pi:
                        continue
                    point = self.elements[element_type]
                    result = helper.log_then_exp(
                        metric=metric, point=point, base_point=base_point
                    )

                    expected = self.group.regularize(point)

                    self.assertAllClose(result, expected, atol=1e-4)

                    if geomstats.tests.tf_backend():
                        break

    @geomstats.tests.np_and_autograd_only
    def test_log_then_exp_left_with_angles_close_to_pi(self):
        """
        Test that the Riemannian left exponential and the
        Riemannian left logarithm are inverse.
        Expect their composition to give the identity function.
        """
        angle_types = self.angles_close_to_pi
        # Canonical inner product on the lie algebra
        for metric in [
            self.metrics_all["left_canonical"],
            self.metrics_all["left_diag"],
        ]:
            for base_point in self.elements.values():
                for element_type in angle_types:
                    point = self.elements_all[element_type]
                    result = helper.log_then_exp(
                        metric=metric, point=point, base_point=base_point
                    )

                    expected = self.group.regularize(point)

                    inv_expected = gs.concatenate([-expected[:3], expected[3:6]])

                    self.assertTrue(
                        gs.allclose(result, expected, atol=1e-6)
                        or gs.allclose(result, inv_expected, atol=1e-6)
                    )

                    if geomstats.tests.tf_backend():
                        break

    @geomstats.tests.np_and_autograd_only
    def test_exp_then_log_left(self):
        """
        Test that the Riemannian left exponential and the
        Riemannian left logarithm are inverse.
        Expect their composition to give the identity function.
        """
        for metric in [
            self.metrics_all["left_canonical"],
            self.metrics_all["left_diag"],
        ]:
            for base_point in self.elements.values():
                for element_type in self.elements:
                    if element_type in self.angles_close_to_pi:
                        continue
                    tangent_vec = self.elements[element_type]
                    result = helper.exp_then_log(
                        metric=metric, tangent_vec=tangent_vec, base_point=base_point
                    )

                    expected = self.group.regularize_tangent_vec(
                        tangent_vec=tangent_vec, base_point=base_point, metric=metric
                    )
                    norm = gs.linalg.norm(expected)
                    atol = ATOL
                    if norm != 0:
                        atol = ATOL * norm
                    self.assertAllClose(result, expected, atol=atol)

                    if geomstats.tests.tf_backend():
                        break

    @geomstats.tests.np_and_autograd_only
    def test_exp_then_log_left_with_angles_close_to_pi(self):
        """
        Test that the Riemannian left exponential and the
        Riemannian left logarithm are inverse.
        Expect their composition to give the identity function.
        """
        angle_types = self.angles_close_to_pi
        # Canonical inner product on the lie algebra
        for metric in [
            self.metrics_all["left_canonical"],
            self.metrics_all["left_diag"],
        ]:
            for base_point in self.elements.values():
                for element_type in angle_types:
                    tangent_vec = self.elements_all[element_type]
                    result = helper.exp_then_log(
                        metric=metric, tangent_vec=tangent_vec, base_point=base_point
                    )

                    expected = self.group.regularize_tangent_vec(
                        tangent_vec=tangent_vec, base_point=base_point, metric=metric
                    )

                    inv_expected = gs.concatenate([-expected[:3], expected[3:6]])

                    self.assertTrue(
                        gs.allclose(result, expected)
                        or gs.allclose(result, inv_expected)
                    )

                    if geomstats.tests.tf_backend():
                        break

    @geomstats.tests.np_and_autograd_only
    def test_log_then_exp_right(self):
        """
        Test that the Riemannian right exponential and the
        Riemannian right logarithm are inverse.
        Expect their composition to give the identity function.
        """
        for metric in [
            self.metrics_all["right_canonical"],
            self.metrics_all["right_diag"],
        ]:
            for base_point in self.elements.values():
                for element_type in self.elements:
                    if element_type in self.angles_close_to_pi:
                        continue
                    point = self.elements[element_type]
                    result = helper.log_then_exp(
                        metric=metric, point=point, base_point=base_point
                    )

                    expected = self.group.regularize(point)
                    expected = gs.cast(expected, gs.float64)
                    norm = gs.linalg.norm(expected)
                    norm = gs.cast(norm, gs.float64)
                    atol = ATOL
                    if norm != 0:
                        atol = ATOL * norm
                    self.assertAllClose(result, expected, atol=atol)

                    if geomstats.tests.tf_backend():
                        break

    @geomstats.tests.np_and_autograd_only
    def test_log_then_exp_right_with_angles_close_to_pi(self):
        """
        Test that the Riemannian right exponential and the
        Riemannian right logarithm are inverse.
        Expect their composition to give the identity function.
        """
        angle_types = self.angles_close_to_pi
        # Canonical inner product on the lie algebra
        for metric in [
            self.metrics_all["right_canonical"],
            self.metrics_all["right_diag"],
        ]:
            for base_point in self.elements.values():
                for element_type in angle_types:
                    point = self.elements_all[element_type]
                    result = helper.log_then_exp(
                        metric=metric, point=point, base_point=base_point
                    )

                    expected = self.group.regularize(point)

                    inv_expected = gs.concatenate([-expected[:3], expected[3:6]])

                    norm = gs.linalg.norm(expected)
                    atol = ATOL
                    if norm != 0:
                        atol = ATOL * norm

                    self.assertTrue(
                        gs.allclose(result, expected, atol=atol)
                        or gs.allclose(result, inv_expected, atol=atol)
                    )

                    if geomstats.tests.tf_backend():
                        break

    @geomstats.tests.np_and_autograd_only
    def test_exp_then_log_right(self):
        """
        Test that the Riemannian left exponential and the
        Riemannian left logarithm are inverse.
        Expect their composition to give the identity function.
        """
        # FIXME
        # for metric in [self.metrics_all['right_canonical'],
        #                self.metrics_all['right_diag']]:
        #     for base_point_type in self.elements:
        #         base_point = self.elements[base_point_type]
        #         for element_type in self.elements:
        #             if element_type in self.angles_close_to_pi:
        #                 continue
        #             tangent_vec = self.elements[element_type]
        #             result = helper.exp_then_log(
        #                                         metric=metric,
        #                                         tangent_vec=tangent_vec,
        #                                         base_point=base_point)

        #             expected = self.group.regularize_tangent_vec(
        #                                         tangent_vec=tangent_vec,
        #                                         base_point=base_point,
        #                                         metric=metric)

        #             if geomstats.tests.tf_backend():
        #                 break

    @geomstats.tests.np_and_autograd_only
    def test_exp_then_log_right_with_angles_close_to_pi(self):
        """
        Test that the Riemannian right exponential and the
        Riemannian right logarithm are inverse.
        Expect their composition to give the identity function.
        """
        # FIXME
        # angle_types = self.angles_close_to_pi
        # # Canonical inner product on the lie algebra
        # for metric in [self.metrics_all['right_canonical'],
        #                self.metrics_all['right_diag']]:
        #     for base_point in self.elements.values():
        #         for element_type in angle_types:
        #             tangent_vec = self.elements_all[element_type]
        #             result = helper.exp_then_log(
        #                                         metric=metric,
        #                                         tangent_vec=tangent_vec,
        #                                         base_point=base_point)

        #             expected = self.group.regularize_tangent_vec(
        #                                         tangent_vec=tangent_vec,
        #                                         base_point=base_point,
        #                                         metric=metric)

        #             inv_expected = gs.concatenate(
        #                 [- expected[:, :3], expected[:, 3:6]],
        #                 axis=1)
        #             norm = gs.linalg.norm(expected)
        #             atol = RTOL
        #             if norm != 0:
        #                 atol = RTOL * norm

        #             if geomstats.tests.tf_backend():
        #                 break

    def test_inner_product_at_identity_vectorization(self):
        n_samples = self.n_samples
        for metric in self.metrics.values():
            one_vector_a = self.group.random_point(n_samples=1)
            one_vector_b = self.group.random_point(n_samples=1)
            n_vector_a = self.group.random_point(n_samples=n_samples)
            n_vector_b = self.group.random_point(n_samples=n_samples)

            result = metric.inner_product(one_vector_a, n_vector_b)
            self.assertAllClose(gs.shape(result), (n_samples,))

            if geomstats.tests.tf_backend():
                break

            result = metric.inner_product(n_vector_a, one_vector_b)
            self.assertAllClose(gs.shape(result), (n_samples,))

            result = metric.inner_product(n_vector_a, n_vector_b)
            self.assertAllClose(gs.shape(result), (n_samples,))

    def test_inner_product_one_base_point_vectorization(self):
        n_samples = self.n_samples
        for metric in self.metrics.values():
            one_base_point = self.group.random_point(n_samples=1)

            one_vector_a = self.group.random_point(n_samples=1)
            one_vector_b = self.group.random_point(n_samples=1)
            n_vector_a = self.group.random_point(n_samples=n_samples)
            n_vector_b = self.group.random_point(n_samples=n_samples)

            result = metric.inner_product(one_vector_a, n_vector_b, one_base_point)
            self.assertAllClose(gs.shape(result), (n_samples,))

            if geomstats.tests.tf_backend():
                break

            result = metric.inner_product(n_vector_a, one_vector_b, one_base_point)
            self.assertAllClose(gs.shape(result), (n_samples,))

            result = metric.inner_product(n_vector_a, n_vector_b, one_base_point)
            self.assertAllClose(gs.shape(result), (n_samples,))

    def test_inner_product_n_base_point_vectorization(self):
        n_samples = self.n_samples
        for metric in self.metrics.values():
            n_base_point = self.group.random_point(n_samples=n_samples)

            one_vector_a = self.group.random_point(n_samples=1)
            one_vector_b = self.group.random_point(n_samples=1)
            n_vector_a = self.group.random_point(n_samples=n_samples)
            n_vector_b = self.group.random_point(n_samples=n_samples)

            result = metric.inner_product(one_vector_a, n_vector_b, n_base_point)
            self.assertAllClose(gs.shape(result), (n_samples,))

            if geomstats.tests.tf_backend():
                break

            result = metric.inner_product(n_vector_a, one_vector_b, n_base_point)
            self.assertAllClose(gs.shape(result), (n_samples,))

            result = metric.inner_product(n_vector_a, n_vector_b, n_base_point)
            self.assertAllClose(gs.shape(result), (n_samples,))

    @geomstats.tests.np_and_autograd_only
    def test_squared_dist_is_symmetric(self):
        for metric in self.metrics.values():
            for point_a in self.elements.values():
                for point_b in self.elements.values():
                    point_a = self.group.regularize(point_a)
                    point_b = self.group.regularize(point_b)

                    sq_dist_a_b = metric.squared_dist(point_a, point_b)
                    sq_dist_b_a = metric.squared_dist(point_b, point_a)

                    self.assertAllClose(sq_dist_a_b, sq_dist_b_a)

                    if geomstats.tests.tf_backend():
                        break

    @geomstats.tests.np_and_autograd_only
    def test_dist_is_symmetric(self):
        for metric in self.metrics.values():
            for point_a in self.elements.values():
                for point_b in self.elements.values():
                    point_a = self.group.regularize(point_a)
                    point_b = self.group.regularize(point_b)

                    dist_a_b = metric.dist(point_a, point_b)
                    dist_b_a = metric.dist(point_b, point_a)
                    self.assertAllClose(dist_a_b, dist_b_a)

                    if geomstats.tests.tf_backend():
                        break

    def test_squared_dist_vectorization(self):
        n_samples = self.n_samples
        for metric_type in self.metrics:
            metric = self.metrics[metric_type]
            point_id = self.group.identity

            one_point_1 = self.group.random_point(n_samples=1)
            one_point_2 = self.group.random_point(n_samples=1)
            one_point_1 = self.group.regularize(one_point_1)
            one_point_2 = self.group.regularize(one_point_2)

            n_point_1 = self.group.random_point(n_samples=n_samples)
            n_point_2 = self.group.random_point(n_samples=n_samples)
            n_point_1 = self.group.regularize(n_point_1)
            n_point_2 = self.group.regularize(n_point_2)

            # Identity and n points 2
            result = metric.squared_dist(point_id, n_point_2)
            self.assertAllClose(gs.shape(result), (n_samples,))

            if geomstats.tests.tf_backend():
                break

            # n points 1 and identity
            result = metric.squared_dist(n_point_1, point_id)
            self.assertAllClose(gs.shape(result), (n_samples,))

            # one point 1 and n points 2
            result = metric.squared_dist(one_point_1, n_point_2)
            self.assertAllClose(gs.shape(result), (n_samples,))

            # n points 1 and one point 2
            result = metric.squared_dist(n_point_1, one_point_2)
            self.assertAllClose(gs.shape(result), (n_samples,))

            # n points 1 and n points 2
            result = metric.squared_dist(n_point_1, n_point_2)
            self.assertAllClose(gs.shape(result), (n_samples,))

    def test_dist_vectorization(self):
        n_samples = self.n_samples
        for metric_type in self.metrics:
            metric = self.metrics[metric_type]
            point_id = self.group.identity

            one_point_1 = self.group.random_point(n_samples=1)
            one_point_2 = self.group.random_point(n_samples=1)
            one_point_1 = self.group.regularize(one_point_1)
            one_point_2 = self.group.regularize(one_point_2)

            n_point_1 = self.group.random_point(n_samples=n_samples)
            n_point_2 = self.group.random_point(n_samples=n_samples)
            n_point_1 = self.group.regularize(n_point_1)
            n_point_2 = self.group.regularize(n_point_2)

            # Identity and n points 2
            result = metric.dist(point_id, n_point_2)
            self.assertAllClose(gs.shape(result), (n_samples,))

            if geomstats.tests.tf_backend():
                break
            # n points 1 and identity
            result = metric.dist(n_point_1, point_id)
            self.assertAllClose(gs.shape(result), (n_samples,))

            # one point 1 and n points 2
            result = metric.dist(one_point_1, n_point_2)
            self.assertAllClose(gs.shape(result), (n_samples,))

            # n points 1 and one point 2
            result = metric.dist(n_point_1, one_point_2)
            self.assertAllClose(gs.shape(result), (n_samples,))

            # n points 1 and n points 2
            result = metric.dist(n_point_1, n_point_2)
            self.assertAllClose(gs.shape(result), (n_samples,))

    def test_geodesic_and_belongs(self):
        initial_point = self.group.random_point()
        initial_tangent_vec = gs.array([2.0, 0.0, -1.0, 0.0, 2.0, 3.0])
        metric = self.metrics_all["left_canonical"]
        geodesic = metric.geodesic(
            initial_point=initial_point, initial_tangent_vec=initial_tangent_vec
        )

        t = gs.linspace(start=0.0, stop=1.0, num=100)
        points = geodesic(t)
        result = gs.all(self.group.belongs(points))
        expected = True
        self.assertAllClose(result, expected)

    def test_geodesic_subsample(self):
        initial_point = gs.array([-1.1, 0.0, 0.99, 10.0, 2.0, 3.0])
        initial_tangent_vec = gs.array([1.0, 0.0, 2.0, 1.0, 1.0, 1.0])
        metric = self.metrics_all["left_canonical"]
        geodesic = metric.geodesic(
            initial_point=initial_point, initial_tangent_vec=initial_tangent_vec
        )
        n_steps = 10
        t = gs.linspace(start=0.0, stop=1.0, num=n_steps + 1)
        points = geodesic(t)

        tangent_vec_step = initial_tangent_vec / n_steps
        for i in range(n_steps + 1):
            point_step = metric.exp(
                tangent_vec=i * tangent_vec_step, base_point=initial_point
            )
            result = point_step
            expected = points[i]
            self.assertAllClose(result, expected)
